# =========================================================
# CreateNetwork.py
# @author Nina Thomsen
# @date 21.12.2022
# @copyright Institut fuer Verkehrsforschung,
#            Deutsches Zentrum fuer Luft- und Raumfahrt
# @brief Creates nodes between edges
# =========================================================

import warnings
from datetime import datetime

import geopandas as gpd
import networkx as nx
import numpy as np
import osmnx as ox
import pandas as pd
from scipy.spatial import cKDTree
from shapely.errors import ShapelyDeprecationWarning
from shapely.geometry import Point, LineString
from sklearn.cluster import KMeans
from tqdm import tqdm


def ckdnearest(gdA, gdB, bcolA, bcolB, n):
    """
    Find n number of nearest nodes from gdA to gdB and calculate their distance

    :param gdA: Nodes A, Point geometry; find n nearest points in gdB for all points in gdA
    :param gdB: Nodes B, Point geometry; contains nodes that are assigned to gdA
    :param bcolA: name of ID column in gdA
    :param bcolB: name of ID column in gdB
    :param n: number of nearest nodes to find per node from gdA
    :type gdA: gpd.GeoDataFrame
    :type gdB: gpd.GeoDataFrame
    :type bcolA: str
    :type bcolB: str
    :type n: int
    :return: pd.DataFrame with all nodes from gdA, the IDs of their n nearest nodes from gdB and their distances
    """
    nA = np.array(list(zip(gdA.geometry.y, gdA.geometry.x)))
    nB = np.array(list(zip(gdB.geometry.y, gdB.geometry.x)))
    btree = cKDTree(nB)
    dist, idx = btree.query(nA, k=n)
    df = pd.DataFrame(columns=[bcolA, 'distance', '{}_near'.format(bcolB)])
    if n > 1:
        for i in range(idx.shape[1]):
            dfi = pd.DataFrame.from_dict({bcolA: gdA[bcolA], 'distance': dist[:, i],  # .astype(int),
                                          '{}_near'.format(bcolB): gdB.reset_index().loc[idx[:, i], bcolB].values})
            df = pd.concat([df, dfi], axis=0)
    else:
        dfi = pd.DataFrame.from_dict({bcolA: gdA[bcolA], 'distance': dist,  # .astype(int),
                                  '{}_near'.format(bcolB): gdB.reset_index().loc[idx, bcolB].values})
        df = pd.concat([df, dfi], axis=0)
    return df


def graph_from_gdf(nodes, edges, node_id='node_id', node_geo='geometry', edge_id='edge_id', edge_from='from_node', edge_to='to_node'):
    """
    Create a MultiDiGraph from nodes and edges in the GeoDataFrame format

    :param nodes: Nodes, include id
    :param edges: Edges, include id, from node id and to node id (directed)
    :param node_id: id column in nodes
    :param node_geo: geometry column in nodes
    :param edge_id: id column in edges
    :param edge_from: from node column in edges
    :param edge_to: to node column in edges
    :type nodes: gpd.GeoDataFrame
    :type edges: gpd.GeoDataFrame
    :type node_id: str
    :type node_geo: str
    :type edge_id: str
    :type edge_from: str
    :type edge_to: str
    :return: MultiDiGraph
    :rtype: MultiDiGraph
    """
    node_graph = nodes.copy()
    node_graph.set_index(node_id, inplace=True)
    node_graph['x'] = node_graph[node_geo].x
    node_graph['y'] = node_graph[node_geo].y
    edge_graph = edges.copy()
    edge_graph['osmid'] = edge_graph[edge_id]
    edge_graph.set_index([edge_from, edge_to, edge_id], inplace=True)

    return ox.utils_graph.graph_from_gdfs(node_graph, edge_graph)


def subordinate_road_length(taz_sub, sub_type="secondary"):
    """
    Calculate the subordinate network length per TAZ for a selected category

    :param taz_sub: TAZ GeoDataFrame
    :param sub_type: highway type for subordinate network
    :return: GeoDataFrame with aggregated network length per TAZ
    :type taz_sub: gpd.GeoDataFrame
    :type sub_type: str
    :rtype: gpd.GeoDataFrame
    """
    taz_sub['length_sub'] = 0
    # iterate over TAZ
    for i, t in tqdm(taz_sub.iterrows()):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)
            sub_edges = ox.geometries.geometries_from_polygon(t['geometry'], tags={"highway": sub_type})
        # remove geometries other than line
        if len(sub_edges.geom_type.unique()) > 1:
            sub_edges = sub_edges[sub_edges.geom_type.isin(["MultiLineString", "LineString"])].copy()

        # calculate length per edge in km and aggregate length per taz
        sub_edges = sub_edges.to_crs(epsg=3035)
        taz_sub.loc[i, "length_sub"] = sub_edges.length.sum() / 1000

    return taz_sub


class Edges:
    """Create network edges from OSM"""

    def __init__(self, country, taz, taz_cn="country", taz_geo="geometry"):
        """

        :param country: code or name of country
        :param taz: GeoDataFrame including the TAZ
        :param taz_geo: Name of geometry column in TAZ layer, default "geometry
        :param taz_cn: Name of column in TAZ layer that defines the country of the TAZ
        :type taz_geo: str
        :type taz_cn: str
        :type taz: gpd.GeoDataFrame
        :type country: str
        """
        self.country = country
        self.edges = gpd.GeoDataFrame()
        self.taz = taz
        self.taz_cn = taz_cn
        self.taz_geo = taz_geo

    def get_polygon(self, taz_id, proj_crs=3035, buffer=2500):
        """
        Define polygon(s) for network extraction; split islands and exclaves and apply buffer to include all streets near the border

        :param taz_id: Name of ID-column in TAZ layer
        :param proj_crs: EPSG number of projected crs for defining buffer in meter, default 3035 (LAEA europe)
        :param buffer: buffer in meter
        :type buffer: float
        :type proj_crs: int
        :type taz_id: str
        :return: GeoDataFrame with polygons for extraction; multiple polygons if there are islands or exclaves
        """
        # start: filter country
        taz_cn = self.taz[self.taz[self.taz_cn] == self.country].copy()
        taz_crs = taz_cn.crs
        # get all separate polygons (islands, exclaves)
        t_e = taz_cn.explode(index_parts=True)

        # find islands belonging to the same TAZ id and create hull
        if len(t_e[taz_id].unique()) > 1:
            isl = t_e.groupby(taz_id)[self.taz_geo].agg('count').reset_index()
            isl = isl[isl[self.taz_geo] > 1]
            # create hull for same TAZ id islands
            for i in isl[taz_id]:
                hull = taz_cn.loc[taz_cn[taz_id] == i].unary_union.convex_hull
                taz_cn.loc[taz_cn[taz_id] == i, self.taz_geo] = hull

        taz_cn_e = taz_cn.dissolve().explode(index_parts=True)
        # apply buffer (include all roads in coastal regions, islands that are not far away and their bridges
        taz_cn_e.to_crs(epsg=proj_crs, inplace=True)
        taz_cn_e[self.taz_geo] = taz_cn_e[self.taz_geo].buffer(buffer)
        taz_cn_e.to_crs(taz_crs, inplace=True)

        # return polygons
        return taz_cn_e

    def get_edges(self, filter_highway, polygons):  # add poly attribute, either none or input
        """
        Get network for selected highway types from OSM for specified country

        :param filter_highway: List of OSM highway types to include; only main types (e.g. motorway), respective
                               link-types (e.g. motorway_link) are included automatically
        :param polygons: polygons of country / region to extract network from, several polygons possible
        :return self.edges includes routable and simplified network
        :type polygons: gpd.GeoDataFrame
        :type filter_highway: list
        """
        filter_highway.extend(["{}_link".format(h) for h in filter_highway])
        fil_str = "|".join(filter_highway)

        filter_nw = '["area"!~"yes"]["highway"~"{}"]["motor_vehicle"!~"no"]["motorcar"!~"no"]["access"!~"private"]["service"!~"parking|parking_aisle|driveway|private|emergency_access"]'.format(
            fil_str)

        # get polygon for country
        polygons = polygons
        n_poly = 0  # number of polygons with road network
        for i, poly in enumerate(polygons[self.taz_geo][:]):
            try:
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)
                    warnings.filterwarnings("ignore", category=FutureWarning, append=True)
                    warnings.filterwarnings("ignore", category=UserWarning, append=True)
                    G = ox.graph_from_polygon(poly, simplify=False, custom_filter=filter_nw, retain_all=True)
                    G = ox.simplify_graph(G)
                    G = ox.simplification.consolidate_intersections(G, tolerance=0.002)
                G = ox.speed.add_edge_speeds(G, fallback=50, precision=0)
                edges = gpd.GeoDataFrame([x[2] for x in G.edges.data()])[["highway", "speed_kph", "geometry"]]
                if i == 0:
                    self.edges = edges.copy()
                else:
                    self.edges = pd.concat([self.edges, edges.reset_index()])
                n_poly += 1
            except:
                print('No edges for polygon {}'.format(i))
        # correct speeds
        self.edges.loc[self.edges["speed_kph"] > 150, "speed_kph"] = 50
        self.edges.crs = "epsg:4326"

        return n_poly

    def set_attributes(self, taz_id, start_id=0):
        """
        Overlay self.edges with TAZ layer to split links at borders and assign TAZ-ID
        Set length, travel time, highway type and ID of self.edges GeoDataFrame

        :param taz_id: Name of ID-column in TAZ layer
        :param start_id: start of IDs for each edge, default 0
        :return: self.edges is updated with necessary attributes
        :type taz_id: str
        :type start_id: int
        """

        self.edges = gpd.overlay(self.edges, self.taz[[taz_id, self.taz_geo]], how='union', keep_geom_type=True)
        self.edges = self.edges.explode(index_parts=True)
        # calculate length in km
        self.edges = self.edges.to_crs(epsg=3035)
        self.edges["length"] = self.edges.length / 1000
        self.edges = self.edges.to_crs(epsg=4326)
        # calculate travel time in s
        self.edges["tt"] = np.round((self.edges["length"] / self.edges["speed_kph"])*(60**2))
        # set ID, type as numeric
        self.edges = self.edges.to_crs(epsg=4326)
        self.edges["ultimo_id"] = np.arange(len(self.edges)) + start_id
        self.edges["type"] = 1
        self.edges.loc[self.edges["highway"] == "trunk", "type"] = 2
        self.edges.loc[self.edges["highway"] == "primary", "type"] = 3
        self.edges.loc[self.edges["highway"] == "motorway_link", "type"] = 11
        self.edges.loc[self.edges["highway"] == "trunk_link", "type"] = 21
        self.edges.loc[self.edges["highway"] == "primary_link", "type"] = 31

    def set_nodes(self, nodes, order_col, nodeid_col="node_id", linkid_col="ultimo_id"):
        """
        Set start and end node for each edge in self.edges based on node GDF
        :param nodes: GeoDataFrame with start and end node per edge
        :param order_col: Name of column with order of nodes per edge in nodes
        :param nodeid_col: Name of column with node ID in nodes, default "node_id"
        :param linkid_col: Name of column with link ID in nodes, default "ultimo_id"
        :return: self.edges includes from and to nodes
        :type linkid_col: str
        :type nodeid_col: str
        :type order_col: str
        :type nodes: gpd.GeoDataFrame
        """
        # merge start nodes
        st_nodes = nodes[nodes[order_col] == 0]
        self.edges = pd.merge(self.edges, st_nodes[[linkid_col, nodeid_col]], left_on="ultimo_id", right_on=linkid_col,
                              how="left")
        self.edges.rename(columns={nodeid_col: "from_node"}, inplace=True)
        # merge start nodes
        en_nodes = nodes[nodes[order_col] == 1]
        self.edges = pd.merge(self.edges, en_nodes[[linkid_col, nodeid_col]], left_on="ultimo_id", right_on=linkid_col,
                              how="left")
        self.edges.rename(columns={nodeid_col: "to_node"}, inplace=True)
        self.edges = self.edges[
            ["ultimo_id", "from_node", "to_node", "type", "nuts_id", "length", "speed_kph", "tt", "geometry"]]

    def connect_subgraphs(self, nodes, edges=None, node_id='node_id', node_geo='geometry', edge_id='ultimo_id', edge_from='from_node', edge_to='to_node'):
        """
        Ensure that the road network is connected, i.e. each node is connected to each other. If there are multiple subgraphs,
        connect the larger subgraphs and remove singular edges.

        :param nodes: network nodes
        :param edges: network edges; if None, use self.edges
        :param node_id: column name with ID in nodes
        :param node_geo: column name with geometry in nodes
        :param edge_id: column name with ID in edges
        :param edge_from: column name with ID of start node in edges
        :param edge_to: column name with ID of end node in edges
        :type nodes: gpd.GeoDataFrame
        :type edges: gpd.GeoDataFrame
        :type node_id: str
        :type node_geo: str
        :type edge_id: str
        :type edge_from: str
        :type edge_to: str
        :return: GeoDataFrames with 1) final network edges and 2) final network nodes
        :rtype: gpd.GeoDataFrame
        """
        if edges is None:
            edges = self.edges

        # get undirected graph
        _graph = graph_from_gdf(nodes, edges, node_id=node_id, node_geo=node_geo, edge_id=edge_id, edge_from=edge_from, edge_to=edge_to)
        _graph_u = ox.utils_graph.get_undirected(_graph)

        # get subgraphs
        SG = [[_graph_u.subgraph(c).copy(), len(c)] for c in sorted(nx.connected_components(_graph_u), key=len, reverse=False)]

        if len(SG) > 1:
            print('Connecting {} sub graphs...'.format(len(SG)))
            # 1 get subgraphs of relevant sizes (more than 5 nodes)
            S = [s[0] for s in SG if s[1] > 5]
            S = [[i, s] for i, s in enumerate(S)]

            # 2 get end nodes per graph
            end_nodes = {}
            for sg in S:
                sg_id = sg[0]
                sg_g = sg[1]
                # get all nodes in sg
                n_sg = [e[:-1] for e in sg_g.edges]
                # get end nodes: all nodes with two or less edges (or minimum number of edges if min>2)
                n_sg = np.unique(n_sg, return_counts=True)
                end_sg = n_sg[0][n_sg[1] <= max(2, min(n_sg[1]))]
                end_nodes.update({sg_id: end_sg})

            # 3 calculate distances between end nodes
            for i in range(len(end_nodes))[:-1]:
                nodes_sg = nodes[nodes[node_id].isin(end_nodes[i])]
                for ii in range(len(end_nodes))[i + 1:]:
                    nodes_sg2 = nodes[nodes[node_id].isin(end_nodes[ii])]
                    # distances between nodes_sg, nodes_sg2
                    if len(nodes_sg2) > 1:
                        dist_ = ckdnearest(nodes_sg, nodes_sg2, node_id, node_id, 2)
                    else:
                        dist_ = ckdnearest(nodes_sg, nodes_sg2, node_id, node_id, 1)
                    # filter minimal distances
                    dist_ = dist_.loc[dist_['distance'] < dist_['distance'].min() * 1.3]
                    # add sg id
                    dist_['S1'] = i
                    dist_['S2'] = ii
                    if (i == 0) & (ii == 1):
                        dist_all = dist_.copy()
                    else:
                        dist_all = pd.concat([dist_all, dist_])

            # 4 connect closest end nodes
            # get min distance per SG
            min_dis_per_sg1 = dist_all.groupby('S1')['distance'].min().reset_index().rename(columns={'S1': 'S'})
            min_dis_per_sg2 = dist_all.groupby('S2')['distance'].min().reset_index().rename(columns={'S2': 'S'})
            min_dis_per_sg = pd.concat([min_dis_per_sg1, min_dis_per_sg2])
            min_dis_per_sg = min_dis_per_sg.groupby('S')['distance'].min().reset_index()
            del min_dis_per_sg1, min_dis_per_sg2
            # sort by distance, ascending
            dist_all.sort_values(by='distance', inplace=True)
            # container for subgraphs
            container = []
            # final road network
            roads_final = edges.copy()
            roads_final[edge_id] = roads_final[edge_id].astype(int)

            for i, row in dist_all.iterrows():
                s1 = row['S1']
                s2 = row['S2']
                min_dis = min_dis_per_sg.loc[min_dis_per_sg['S'].isin([s1, s2]), 'distance'].min()
                connect = False
                # check if s1 is already in container
                if s1 in container:
                    # if s2 not in container, connect
                    if s2 not in container:
                        connect = True
                    # if s2 also in container, connect only if distance in low
                    elif row['distance'] < min_dis * 1.1:
                        connect = True
                # connect if s1 not in container
                else:
                    connect = True

                # connect if True
                if connect:
                    # add s1, s2 to container and create line between nodes
                    container.extend([s1, s2])
                    with warnings.catch_warnings():
                        warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)
                        # create lines in both directions
                        node_from = nodes.loc[nodes[node_id] == row[node_id], node_geo]
                        node_to = nodes.loc[nodes[node_id] == row['{}_near'.format(node_id)], node_geo]
                        line = LineString([[node_from.x, node_from.y], [node_to.x, node_to.y]])
                        # add to road network
                        id_ = roads_final[edge_id].max()
                        roads_final.loc[len(roads_final) + 1] = [id_ + 1, row[node_id], row['{}_near'.format(node_id)], 9, '{}000'.format(self.country), 0,
                                                                 50, 0, line]
                        roads_final.loc[len(roads_final) + 1] = [id_ + 2, row['{}_near'.format(node_id)], row[node_id], 11, '{}000'.format(self.country), 0,
                                                                 50, 0, line]

            # 5 remove single edges and small subgraphs
            S_del = [_graph_u.subgraph(c).copy() for c in nx.connected_components(_graph_u) if len(c) <= 5]
            # get nodes and edges from RU_del
            nodes_del = []
            edges_del = []
            for sg in S_del:
                # get all nodes and edges in sg
                n_sg = np.unique([e[:-1] for e in (sg.edges)])
                e_sg = np.unique([e[2] for e in (sg.edges)])
                nodes_del.extend(n_sg)
                edges_del.extend(e_sg)
            roads_final = roads_final.loc[~roads_final[edge_id].isin(edges_del)]
            nodes_final = nodes.loc[~nodes[node_id].isin(nodes_del)]

        else:
            roads_final, nodes_final = edges, nodes

        roads_final.crs=4326
        nodes_final.crs = 4326

        return roads_final, nodes_final


class Nodes:
    """Create nodes at the ends of edges"""

    def __init__(self, edges):
        """
        :param edges: GeoDataFrame with LineString-Objects
        :type edges: gpd.GeoDataFrame
        """
        self.edges = edges
        self.nodes = gpd.GeoDataFrame()
        self.nodes_unique = gpd.GeoDataFrame()

    def init_nodes(self):
        """init node GeoDataFrame"""
        self.nodes = gpd.GeoDataFrame(columns=['LinkID', 'order', 'geometry'])

    def create_nodes(self, id_col, geom_col="geometry"):
        """
        Create nodes at the ends of each LineString in edges
        !! only works with LineStrings, not with MultiLineString
        !! use gpd.GeoDataFrame.explode() to transform MultiLineString to LineString

        :param id_col: name of column with unique identifier in edges GDF
        :param geom_col: name of geometry column in edges GDF, default "geometry"
        :return GeoDataFrame self.nodes
        :type geom_col: str
        :type id_col: str
        """
        self.init_nodes()
        LinkId_list = self.edges[id_col].to_list()
        coords_list = [line.coords for i, line in enumerate(self.edges[geom_col])]
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)
            for LinkId, coords in zip(LinkId_list, coords_list):
                index_num = self.nodes.shape[0]
                # 1st point
                self.nodes.loc[index_num, 'geometry'] = Point(coords[0])
                self.nodes.loc[index_num, 'LinkID'] = LinkId
                self.nodes.loc[index_num, 'order'] = 0
                # last point
                self.nodes.loc[index_num + 1, 'geometry'] = Point(coords[-1])
                self.nodes.loc[index_num + 1, 'LinkID'] = LinkId
                self.nodes.loc[index_num + 1, 'order'] = 1
        self.nodes.set_geometry('geometry', inplace=True)
        self.nodes.crs = 4326

    def remove_duplicates(self):
        """
        Remove duplicate nodes (i.e. at the same coordinates)

        --> result: GeoDataFrame self.nodes.unique without duplicates

        :return: None
        """
        self.nodes['xy'] = [str(list(p.coords)) for p in self.nodes['geometry']]
        # count links per point
        dup = pd.DataFrame(self.nodes.groupby(['xy'])['LinkID'].agg('count').reset_index())
        # transform to point data
        dup['_xy'] = [x[2:-2] for x in dup['xy']]
        dup[["x", "y"]] = [x.split(", ") for x in dup['_xy']]
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)
            dup['geometry'] = [Point(float(x), float(y)) for x, y in zip(dup['x'], dup['y'])]
        dup.rename(columns={'LinkID': 'num_links'}, inplace=True)
        self.nodes_unique = gpd.GeoDataFrame(dup[["xy", "num_links", "geometry"]], geometry="geometry")
        self.nodes_unique.crs = self.nodes.crs

    def set_node_id(self, id_col="node_id", start=0):
        """
        Set ID per unique node and assign this ID to all nodes

        --> result: GeoDataFrame of self.nodes, self.nodes_unique with ID column

        :param id_col: name of ID column for nodes, default "node_id"
        :param start: start of IDs for each node, default 0
        :return: None
        :type start: int
        :type id_col: str
        """
        self.nodes_unique[id_col] = np.arange(len(self.nodes_unique)) + start
        # join unique ID to all nodes
        self.nodes = pd.merge(self.nodes, self.nodes_unique[["xy", id_col]], how="left", on="xy")
        self.nodes = self.nodes[[id_col, "LinkID", "order", "geometry"]]
        self.nodes_unique = self.nodes_unique[[id_col, "num_links", "geometry"]]


class Connectors:
    """
    Connect the TAZ to the road network: find possible connector locations and identify corresponding nodes
    """

    def __init__(self):
        """


        """
        self.connectors = gpd.GeoDataFrame()

    def find_connector_locations(self, taz, pop, taz_id="nuts_id", taz_geom="geom", epsg_pop=4326):
        """
        Identify suitable connector locations per TAZ depending on population density

        --> result is self.connectors as GeoDataFrame with connector locations; Point layer

        :param epsg_pop: CRS of population dataframe, default 4326
        :param taz: GeoDataFrame with the TAZ
        :param taz_geom: name of geometry column in TAZ GDF
        :param taz_id: name of ID column for TAZ IDs
        :param pop: GeoDataFrame of points with population density

        :type epsg_pop: int
        :type taz: gpd.GeoDataFrame
        :type taz_geom: str
        :type taz_id: str
        :type pop: gpd.GeoDataFrame

        :return: None
        :rtype:
        """
        cols = ["nuts_id", "c_n", "weight", "geometry"]
        return_con = gpd.GeoDataFrame(columns=cols)

        taz = taz.to_crs(epsg=3035)
        taz['area'] = taz.area / 1000 ** 2
        taz["numcon"] = 3
        taz.loc[taz["area"] > 1000, "numcon"] = 4
        taz.loc[taz["area"] > 7500, "numcon"] = 5
        taz.loc[taz["area"] > 15000, "numcon"] = 6

        # overlay pop and taz
        taz = taz.to_crs(epsg=epsg_pop)
        pop_taz = gpd.overlay(pop, taz[[taz_id, taz_geom]])

        for t in taz[taz_id].unique():
            pop_t = pop_taz[pop_taz[taz_id] == t]
            if len(pop_t) > 0:
                n_c = int(taz.loc[taz[taz_id] == t, "numcon"])
                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", category=UserWarning)
                    kmeans = KMeans(init="random", n_clusters=n_c, n_init=10, max_iter=300, random_state=42)
                    kmeans.fit(list(zip(pop_t.geometry.x, pop_t.geometry.y)))
                # assign pop to cluster
                pop_t = pd.concat([pop_t.reset_index(), pd.Series(kmeans.labels_, name="c_n")], axis=1)
                # create GeoDataFrame with weights and coordinates of centers (connectors)
                conn = gpd.GeoDataFrame(pop_t.groupby("c_n")['VALUE'].apply(sum).reset_index(),
                                        geometry=[Point(xy) for xy in kmeans.cluster_centers_])
                conn['weight'] = conn['VALUE'] / conn['VALUE'].sum()
                conn['nuts_id'] = t
                conn = conn[cols]
                return_con = pd.concat([return_con, conn])

        self.connectors = return_con
        self.connectors.crs = epsg_pop
        self.connectors['c_n'] = self.connectors.reset_index().index

    def identify_connector_nodes(self, nodes, node_no="node_id", node_geom="geometry", zone="nuts_id",
                                 weight="weight", country_check=False):
        """
        Move connector locations to network nodes

        :param nodes: GeoDataFrame Points; network nodes
        :param node_no: column name with node identifyer
        :param node_geom: name of geometry column in nodes GDF
        :param zone: column name with zone identifyer
        :param weight: column name with weight of connector
        :return: GeoDataFrame with connector nodes

        :type nodes: gpd.GeoDataFrame
        :type node_no: str
        :type node_geom: str
        :type zone: str
        :type weight: str
        :rtype: gpd.GeoDataFrame
        """
        con = self.connectors.copy()

        # find one node per connector location
        distances = ckdnearest(con, nodes, 'c_n', node_no, 1)
        output = con[[zone, 'c_n', weight]].merge(distances, on='c_n')
        output.rename(columns={'{}_near'.format(node_no): node_no}, inplace=True)

        # check for duplicates
        con_nodes = output[[node_no, zone, 'c_n']].groupby([zone, node_no], as_index=False).count()

        if len(con_nodes) != len(output):
            print('Duplicates detected!')
            output["distprod"] = output['distance'] * 1 / output[weight]
            dup_no = con_nodes.loc[con_nodes['c_n'] > 1][node_no].tolist()
            dup = output[output[node_no].isin(dup_no)].copy()
            min_dis = dup.groupby([zone, node_no])['distprod'].transform('min').unique()
            delete = dup.loc[~dup['distprod'].isin(min_dis), 'c_n'].tolist()
            print(len(delete), sum(output.loc[output['c_n'].isin(delete), weight]))
            output = output[~output['c_n'].isin(delete)].copy()
            # correct weights: add deleted weight to remaining connector (duplicates only removed for same TAZ)
            delete_weights = dup.groupby([zone, node_no])[weight].apply(sum).reset_index()
            delete_weights.columns = [zone, node_no, "newweight"]
            output = output.merge(delete_weights, on=[zone, node_no], how="left")
            output.loc[~output['newweight'].isna(), weight] = output['newweight']
            print(sum(output[weight]))
        else:
            print('No duplicates detected!')

        # check if connector nodes are within the same country
        if country_check:
            output['cn_connode'] = output[zone].str[:2]
            move_cons = nodes[[node_no, country_check]].merge(output[[node_no, 'c_n', 'cn_connode']], on=node_no, how="right")
            move_cons = move_cons.loc[move_cons['cn_connode'] != move_cons[country_check]]
            move_cons = move_cons.merge(self.connectors[['c_n', 'geometry']], on='c_n', how='left')
            move_cons = move_cons.set_geometry('geometry')
            move_cons.crs = 4326
            move_cons.drop(columns=[node_no, country_check], inplace=True)
            # move connector nodes in move_cons
            print('Move {} connectors from foreign nodes...'.format(len(move_cons)))
            for c in move_cons['cn_connode'].unique():
                nodes_c = nodes[nodes[country_check]==c]
                dist_c = ckdnearest(move_cons[move_cons['cn_connode']==c], nodes_c, 'c_n', node_no, 1)
                dist_c.rename(columns={'{}_near'.format(node_no): node_no}, inplace=True)
                # replace node_no in output
                output.loc[output['c_n'].isin(dist_c['c_n']), node_no] = dist_c[node_no].tolist()
            output.drop(columns=['cn_connode'], inplace=True)

        # create GDF with correct connector nodes
        connodes = nodes[[node_no, node_geom]].merge(output[[node_no, zone, 'c_n', weight]], on=node_no, how="right")
        connodes.set_geometry(node_geom, inplace=True)

        return connodes


class Ferries:
    """
    Create connections over water between islands and main land, using ferry routes and bridges as reference
    """

    def __init__(self, taz, scope=None, taz_cn='cntr_code', taz_geo='geometry'):
        """

        :param taz: GeoDataFrame with TAZ cells
        :param scope: region of taz to be considered; default None includes all taz, str with country code reduces taz to country, list with country codes reduces taz to countries in list
        :param taz_cn: name of column with country code in taz
        :param taz_geo: name of geometry column in taz

        :type taz: gpd.GeoDataFrame
        :type scope: str or list
        :type taz_cn: str
        :type taz_geo: str
        """
        self.taz = taz
        self.scope = scope
        self.taz_cn = taz_cn
        self.taz_geo = taz_geo
        # outputs
        if scope is None:
            self.region = taz
        elif type(scope) is str:
            self.region = taz[taz[taz_cn] == scope]
        else:
            self.region = taz[taz[taz_cn].isin(scope)]
        self.ferry = gpd.GeoDataFrame()
        self.nodes = gpd.GeoDataFrame()

    def find_ferry_routes(self, buffer_water=0.05):
        """
        Extract ferry routes from OSM for water body around region (defined by scope in __init__)

        --> result: self.ferry and self.nodes include ferry routes and their end nodes

        :param buffer_water: buffer to use around land mass to extract water polygon, in degrees
        :type buffer_water: float
        :return: None
        """
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)
            warnings.filterwarnings("ignore", category=FutureWarning, append=True)
            warnings.filterwarnings("ignore", category=UserWarning, append=True)
            # create polygon of water body for region
            hull = self.region.unary_union.convex_hull
            water = hull.difference(self.region[self.taz_geo].buffer(buffer_water).unary_union)
            # get ferry routes and bridges with roads for water body
            print(". . . extracting bridges and ferries from OSM")
            self.ferry = ox.geometries.geometries_from_polygon(water, tags={"route": "ferry", "highway": ["motorway", "trunk"]})
            print(". . . bridges and ferries extracted!")
            # remove non-LineString geometries
            if len(self.ferry.geom_type.unique()) > 1:
                self.ferry = self.ferry[self.ferry.geom_type.isin(["MultiLineString", "LineString"])].copy()
            # filter only ferries with motor vehicle or bridges with highway attribute
            self.ferry = self.ferry[(~self.ferry['highway'].isna()) | (self.ferry['motor_vehicle'] == "yes")].copy()
            # set attributes and reduce to relevant columns
            self.ferry['id'] = list(range(len(self.ferry)))
            self.ferry = self.ferry.reset_index()
            self.ferry['type'] = 9  # ferry
            self.ferry.loc[~self.ferry['highway'].isna(), 'type'] = 8  # bridge
            self.ferry = self.ferry[['id', 'geometry', 'type']].copy()
            # create ferry end nodes
            nodes = Nodes(self.ferry)
            nodes.create_nodes('id')
            nodes.remove_duplicates()
            nodes.set_node_id()
            self.nodes = nodes.nodes.set_geometry('geometry')
            self.nodes.crs = 4326

    def ferry_national(self, region_id='id', ferry_buffer=0.01):
        """
        Create GeoDataFrames of ferry routes and nodes with a country; filter other relations out of self.ferry and self.nodes

        :param region_id: name of column with identifier
        :param ferry_buffer: buffer to use to look for end nodes in coast region
        :return: GeoDataFrame ferry_routes with national ferry routes and GeoDataFrame ferry_nodes with respective nodes
        :type region_id: str
        :type ferry_buffer: float
        :rtype: gpd.GeoDataFrame
        """
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)
            warnings.filterwarnings("ignore", category=FutureWarning, append=True)
            warnings.filterwarnings("ignore", category=UserWarning, append=True)
            # find nodes in border area
            self.region.to_crs(self.nodes.crs, inplace=True)
            reg_buf = self.region.copy()
            reg_buf[self.taz_geo] = reg_buf[self.taz_geo].buffer(ferry_buffer)
            border_ferry = gpd.overlay(self.nodes, reg_buf[[region_id, self.taz_geo]])
            # find end nodes within two different cells
            n_g = border_ferry.groupby('LinkID')['order', region_id].nunique().reset_index()
            filter_ferry = list(n_g.loc[(n_g['order'] > 1) & (n_g[region_id] > 1), 'LinkID'])
            # final result
            ferry_routes = self.ferry[self.ferry['id'].isin(filter_ferry)].copy()
            ferry_nodes = self.nodes[self.nodes['LinkID'].isin(filter_ferry)].copy()

        return ferry_routes, ferry_nodes

    def ferry_international(self, ferry_buffer=0.01):
        """
        Create GeoDataFrame of ferry routes and nodes between different countries, filter national relations out of self.ferry and self.nodes

        :param ferry_buffer: buffer to use to look for end nodes in coast region
        :return: GeoDataFrame ferry_routes with international ferry routes and GeoDataFrame ferry_nodes with respective nodes

        :type ferry_buffer: float
        :rtype: gpd.GeoDataFrame
        """
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)
            warnings.filterwarnings("ignore", category=FutureWarning, append=True)
            warnings.filterwarnings("ignore", category=UserWarning, append=True)
            # find nodes in border area
            self.region.to_crs(self.nodes.crs, inplace=True)
            reg_buf = self.region.copy()
            reg_buf[self.taz_geo] = reg_buf[self.taz_geo].buffer(ferry_buffer)
            border_ferry = gpd.overlay(self.nodes, reg_buf[[self.taz_cn, self.taz_geo]])
            # find end nodes within two different countries
            n_g = border_ferry.groupby('LinkID')['order', self.taz_cn].nunique().reset_index()
            filter_ferry = list(n_g.loc[(n_g['order'] > 1) & (n_g[self.taz_cn] > 1), 'LinkID'])
            # final result
            ferry_routes = self.ferry[self.ferry['id'].isin(filter_ferry)].copy()
            ferry_nodes = self.nodes[self.nodes['LinkID'].isin(filter_ferry)].copy()

        return ferry_routes, ferry_nodes

    def ferry_to_road(self, ferry_routes, ferry_nodes, network_roads, network_nodes, ferry_buffer_m=2500, speed_ferry=30,
                      roads_id="ultimo_id", roads_fr="from_node", roads_to="to_node", node_id="node_id"):
        """
        Connect ferry end nodes to road network and add as links (direct lines)

        :param ferry_routes: GeoDataFrame with ferry routes (lines)
        :param ferry_nodes: GeoDataFrame with ferry nodes (points)
        :param network_roads: GeoDataFrame with road network (lines)
        :param network_nodes: GeoDataFrame with unique road network nodes (points)
        :param ferry_buffer_m: buffer size around ferry nodes for finding connecting roads in meter
        :param speed_ferry: speed on ferry link in kph
        :param roads_id: name of column with road identifier in network_roads
        :param roads_fr: name of column with from node in network_roads
        :param roads_to: name of column with to node in network_roads
        :param node_id: name of column with node identifier in network_nodes
        :return: GeoDataFrame of road network, connected along ferry routes

        :type ferry_routes: gpd.GeoDataFrame
        :type ferry_nodes: gpd.GeoDataFrame
        :type network_roads: gpd.GeoDataFrame
        :type network_nodes: gpd.GeoDataFrame
        :type ferry_buffer_m: float
        :type speed_ferry: float
        :type roads_id: str
        :type roads_fr: str
        :type roads_to: str
        :type node_id: str
        :rtype: gpd.GeoDataFrame
        """
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)
            warnings.filterwarnings("ignore", category=FutureWarning, append=True)
            warnings.filterwarnings("ignore", category=UserWarning, append=True)
            # find roads within buffer of ferry nodes
            ferry_buffer = ferry_nodes.copy()
            ferry_buffer.to_crs(epsg=3035, inplace=True)
            ferry_buffer['geometry'] = ferry_buffer['geometry'].buffer(ferry_buffer_m)
            ferry_buffer.to_crs(epsg=4326, inplace=True)
            border_roads = gpd.overlay(network_roads, ferry_buffer)

            # iterate over ferry routes, connect with closest road node in both directions
            lines = gpd.GeoDataFrame(columns=network_roads.columns)
            id_st = int(network_roads[roads_id].max())

            for ferry in ferry_routes['id'].unique():
                roads_ferry = border_roads.loc[border_roads['LinkID'] == ferry]
                nodes_ferry = ferry_nodes.loc[ferry_nodes['LinkID'] == ferry]
                if len(roads_ferry) == 0:
                    print('No road connection for ferry route {}'.format(ferry))
                else:
                    # start node of ferry
                    roads_ferry_0 = roads_ferry[roads_ferry['order'] == 0]
                    node_ferry_0 = nodes_ferry[nodes_ferry['order'] == 0]
                    # end node of ferry
                    roads_ferry_1 = roads_ferry[roads_ferry['order'] == 1]
                    node_ferry_1 = nodes_ferry[nodes_ferry['order'] == 1]
                    if (len(roads_ferry_0) > 0) & (len(roads_ferry_1) > 0):
                        # length ferry route
                        len_ferry = ferry_routes[ferry_routes['id'] == ferry].copy()
                        len_ferry.to_crs(epsg=3035, inplace=True)
                        len_ferry['len'] = len_ferry.length / 1000
                        len_ferry = float(len_ferry['len'])
                        # type (ferry=9, bridge=8)
                        type_ferry = int(ferry_routes.loc[ferry_routes['id'] == ferry, 'type'])
                        if type_ferry == 8:
                            speed = 60
                        else:
                            speed = speed_ferry

                        # connect roads at start and end
                        # start from nodes 1-0
                        nodes_0f = roads_ferry_0[[roads_id, roads_fr]].rename(columns={roads_fr: node_id})
                        # find closest node to ferry node
                        nodes_0f = nodes_0f.merge(network_nodes[[node_id, 'geometry']], how='left',
                                                  left_on=node_id, right_on=node_id)
                        nodes_0f = nodes_0f.set_geometry('geometry')
                        node_0f = ckdnearest(node_ferry_0, nodes_0f, node_id, node_id, 1)
                        node_0f = nodes_0f[nodes_0f[node_id] == int(node_0f['{}_near'.format(node_id)])]
                        # end to nodes 1-0
                        nodes_1t = roads_ferry_1[[roads_id, roads_to]].rename(columns={roads_to: node_id})
                        # find closest node to ferry node
                        nodes_1t = nodes_1t.merge(network_nodes[[node_id, 'geometry']], how='left',
                                                  left_on=node_id, right_on=node_id)
                        nodes_1t = nodes_1t.set_geometry('geometry')
                        node_1t = ckdnearest(node_ferry_1, nodes_1t, node_id, node_id, 1)
                        node_1t = nodes_1t[nodes_1t[node_id] == int(node_1t['{}_near'.format(node_id)])]
                        node_pair_10 = [node_1t.iloc[0], node_0f.iloc[0]]
                        # start to nodes 0-1
                        nodes_0t = roads_ferry_0[[roads_id, roads_to]].rename(columns={roads_to: node_id})
                        # find closest node to ferry node
                        nodes_0t = nodes_0t.merge(network_nodes[[node_id, 'geometry']], how='left',
                                                  left_on=node_id, right_on=node_id)
                        nodes_0t = nodes_0t.set_geometry('geometry')
                        node_0t = ckdnearest(node_ferry_0, nodes_0t, node_id, node_id, 1)
                        node_0t = nodes_0t[nodes_0t[node_id] == int(node_0t['{}_near'.format(node_id)])]
                        # end from nodes 0-1
                        nodes_1f = roads_ferry_1[[roads_id, roads_fr]].rename(columns={roads_fr: node_id})
                        # find clostest node to ferry node
                        nodes_1f = nodes_1f.merge(network_nodes[[node_id, 'geometry']], how='left',
                                                  left_on=node_id, right_on=node_id)
                        nodes_1f = nodes_1f.set_geometry('geometry')
                        node_1f = ckdnearest(node_ferry_1, nodes_1f, node_id, node_id, 1)
                        node_1f = nodes_1f[nodes_1f[node_id] == int(node_1f['{}_near'.format(node_id)])]
                        node_pair_01 = [node_0t.iloc[0], node_1f.iloc[0]]
                        pairs = [node_pair_01, node_pair_10]
                        # create line
                        for l in pairs:
                            # get coordinates and create LineString
                            node_from = l[0]['geometry']
                            node_to = l[1]['geometry']
                            line = LineString([[node_from.x, node_from.y], [node_to.x, node_to.y]])
                            # add to Lines GDF
                            lines.loc[len(lines) + 1] = [id_st, l[0][node_id], l[1][node_id], type_ferry, 'FERRY',
                                                         len_ferry, speed, (len_ferry / speed) * 60 ** 2, line]
                            id_st += 1
                    else:
                        print('no start / end connection for ferry route {}'.format(ferry))

        network_roads = pd.concat([network_roads, lines])

        return network_roads


class CombineNetworks:
    """
    Merge road networks of multiple countries / regions to create a routable international network
    """

    def __init__(self, taz, roads, taz_cn='cntr_code'):
        """

        :param taz: TAZ for regions
        :param roads: road network covering all regions
        :param taz_cn: column name for country in taz

        :type taz: gpd.GeoDataFrame
        :type roads: gpd.GeoDataFrame
        :type taz_cn: str
        """
        self.taz = taz
        self.taz_cn = taz_cn
        self.network = roads
        self.network_int = gpd.GeoDataFrame()
        self.border_roads = gpd.GeoDataFrame()
        self.borders = gpd.GeoDataFrame()
        self.countries = taz[taz_cn].unique()
        self.dict_borders = {}

    def find_borders(self, buffer=2500, taz_geo='geometry'):
        """
        Create polygons with a defined buffer around the borders

        --> result: self.border contains border polygons

        :param buffer: buffer width in m
        :param taz_geo: name of geometry column in TAZ
        :return: None

        :type buffer: float
        :type taz_geo: str
        """
        # dissolve zones per country
        taz_dis = self.taz.dissolve(by=self.taz_cn, aggfunc="sum").reset_index()
        # set buffer around polygon borders (width of border polygons)
        taz_dis.to_crs(epsg=3035, inplace=True)
        taz_dis[taz_geo] = taz_dis[taz_geo].buffer(buffer / 2)
        # find borders for each country
        gdf_borderbuffer = gpd.GeoDataFrame()

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)
            warnings.filterwarnings("ignore", category=FutureWarning, append=True)
            warnings.filterwarnings("ignore", category=UserWarning, append=True)

            for country in self.countries:

                country_ = taz_dis[taz_dis[self.taz_cn] == country]
                country_ = country_[taz_geo].buffer(1)

                for index, row in taz_dis.iterrows():

                    if row[self.taz_cn] != country:
                        intersec = country_.intersection(row[taz_geo].buffer(1))
                        if not intersec.values.is_empty[0]:
                            new_border = gpd.GeoDataFrame(geometry=intersec)
                            new_border[['country1', 'country2']] = [country, row[self.taz_cn]]
                            gdf_borderbuffer = pd.concat([gdf_borderbuffer, new_border])

            # merge borders create and set geometry
            borders = pd.merge(gdf_borderbuffer, gdf_borderbuffer, left_on=['country1', 'country2'],
                               right_on=['country2', 'country1'])
            borders['geometry'] = [r['geometry_x'].union(r['geometry_y']) for i, r in borders.iterrows()]
            borders['border'] = borders['country1_x'] + borders['country1_y']
            borders = borders.set_geometry('geometry')
            borders.crs = 3035
            # remove duplicates
            borders['abc'] = ["".join(sorted(row['border'])) for i, row in borders.iterrows()]
            borders.drop_duplicates(subset="abc", inplace=True)
            # finalize
            borders = borders[['border', 'country1_x', 'country1_y', 'geometry']]
            borders.to_crs(epsg=4326, inplace=True)

            self.borders = borders

    def get_border_roads(self, roads_cn="cn"):
        """
        Find all roads within the border buffer

        --> result: self.border_roads as GDF with all roads within border buffer

        :param roads_cn: column name with country attribute in roads GDF
        :return: None
        :type roads_cn: str
        """
        if self.borders.crs == self.network.crs:
            self.border_roads = gpd.overlay(self.network, self.borders, how='union', keep_geom_type=True)
            # remove all streets that are not within the border
            self.border_roads = self.border_roads[~self.border_roads['border'].isna()].copy()
            # remove streets that are assigned to a wrong border
            self.border_roads = self.border_roads[(self.border_roads[roads_cn] == self.border_roads['country1_x']) |
                                                  (self.border_roads[roads_cn] == self.border_roads['country1_y'])]
        else:
            print("CRS not matching!\n    Borders {}\n    Roads {}".format(self.borders.crs, self.border_roads.crs))

    def connect_border_roads(self, nodes_int, id_st=900000000, roads_cn='cn', roads_type='type', filter_types=None,
                             roads_id="ultimo_id", roads_fr="from_node", roads_to="to_node", node_id="node_id",
                             node_geo="geometry"):
        """
        Create new directed lines between the nearest end nodes of each country at the border

        --> result: self.network_int as GDF with connected network,
        --> result: self.dict_borders as dictionary with number of connected roads

        :param nodes_int: contains nodes of network
        :param id_st: start ID for new lines
        :param roads_cn: column name for country in self.network
        :param roads_type: column name for road type in self.network
        :param filter_types: types of roads to be included for international connections
        :param roads_id: column name for id in self.network
        :param roads_fr: column name for the start node in self.network
        :param roads_to: column name for the destination node in self.network
        :param node_id: column name for id in nodes
        :param node_geo: column name for geometry in nodes
        :return: None

        :type nodes_int: gpd.GeoDataFrame
        :type id_st: int
        :type roads_cn: str
        :type roads_type: str
        :type filter_types: list
        :type roads_id: str
        :type roads_fr: str
        :type roads_to: str
        :type node_id: str
        :type node_geo: str
        """
        if filter_types is None:
            filter_types = [1, 2]
        lines = gpd.GeoDataFrame(columns=self.network.columns)

        id_st = id_st
        # iterate over borders
        print("Start connecting border roads at {}".format(datetime.now()))
        for b in self.borders['border'].unique():
            # find all fitting edges for both countries
            edges_b1 = self.border_roads[(self.border_roads[roads_cn] == b[:2]) & (self.border_roads['border'] == b) &
                                         (self.border_roads[roads_type].isin(filter_types))]
            edges_b2 = self.border_roads[(self.border_roads[roads_cn] == b[2:]) & (self.border_roads['border'] == b) &
                                         (self.border_roads[roads_type].isin(filter_types))]
            # check if there are edges in both countries
            if (len(edges_b1) > 0) & (len(edges_b2) > 0):
                # get end nodes in both countries
                nodes_b11 = edges_b1[[roads_id, roads_fr]].rename(columns={roads_fr: node_id})
                nodes_b11['dir'] = 'from'
                nodes_b12 = edges_b1[[roads_id, roads_to]].rename(columns={roads_to: node_id})
                nodes_b12['dir'] = 'to'
                nodes_b1 = pd.concat([nodes_b11, nodes_b12]).groupby('node_id')[roads_id].agg('count').reset_index()
                if len(nodes_b1[nodes_b1[roads_id] == nodes_b1[roads_id].min()]) == 1:
                    nodes_b1 = nodes_b1[nodes_b1[roads_id] == nodes_b1[roads_id].min()+1]
                else:
                    nodes_b1 = nodes_b1[nodes_b1[roads_id] == nodes_b1[roads_id].min()]
                nodes_b21 = edges_b2[[roads_id, roads_fr]].rename(columns={roads_fr: node_id})
                nodes_b21['dir'] = 'from'
                nodes_b22 = edges_b2[[roads_id, roads_to]].rename(columns={roads_to: node_id})
                nodes_b22['dir'] = 'to'
                nodes_b2 = pd.concat([nodes_b21, nodes_b22]).groupby('node_id')[roads_id].agg('count').reset_index()
                if len(nodes_b2[nodes_b2[roads_id] == nodes_b2[roads_id].min()]) == 1:
                    nodes_b2 = nodes_b2[nodes_b2[roads_id] == nodes_b2[roads_id].min()+1]
                else:
                    nodes_b2 = nodes_b2[nodes_b2[roads_id] == nodes_b2[roads_id].min()]
                nodes_dir = pd.concat([nodes_b11, nodes_b12, nodes_b21, nodes_b22])
                # get coordinates
                nodes_b1 = nodes_b1.merge(nodes_int[[node_id, node_geo]], how='left', on=node_id)
                nodes_b1 = nodes_b1.set_geometry(node_geo)
                nodes_b2 = nodes_b2.merge(nodes_int[[node_id, node_geo]], how='left', on=node_id)
                nodes_b2 = nodes_b2.set_geometry(node_geo)
                # get the two nearest nodes for each node
                nearest_b1 = ckdnearest(nodes_b1, nodes_b2, node_id, node_id, 2)
                nearest_b2 = ckdnearest(nodes_b2, nodes_b1, node_id, node_id, 2)
                # find nearest node pairs (nearest nodes match) and remove duplicate connections
                pairs_b = nearest_b1.merge(nearest_b2, left_on=[node_id, '{}_near'.format(node_id)],
                                           right_on=['{}_near'.format(node_id), node_id],
                                           how='inner')[['{}_x'.format(node_id), '{}_y'.format(node_id), 'distance_x']]
                # find correct direction
                pairs_b = pairs_b.merge(nodes_dir[[node_id, 'dir']], left_on='{}_x'.format(node_id), right_on=node_id,
                                        how='left')
                pairs_b.drop(columns=[node_id], inplace=True)
                pairs_b = pairs_b.merge(nodes_dir[[node_id, 'dir']], left_on='{}_y'.format(node_id), right_on=node_id,
                                        how='left')
                pairs_b.drop(columns=[node_id], inplace=True)
                pairs_b = pairs_b.loc[:, ~pairs_b.columns.duplicated()]
                pairs_b = pairs_b[pairs_b['dir_x'] != pairs_b['dir_y']]
                pairs_b['node_from'] = [
                    row['{}_x'.format(node_id)] if row['dir_x'] == 'to' else row['{}_y'.format(node_id)] for i, row in
                    pairs_b.iterrows()]
                pairs_b['node_to'] = [
                    row['{}_x'.format(node_id)] if row['dir_x'] == 'from' else row['{}_y'.format(node_id)] for i, row in
                    pairs_b.iterrows()]
                # remove duplicate connections (if a node is paired with multiple other nodes)
                pairs_b = pairs_b.sort_values(by='distance_x')
                pairs_b = pairs_b.drop_duplicates(subset='node_from')
                pairs_b = pairs_b.drop_duplicates(subset='node_to')
                pairs_b = [[row['node_from'], row['node_to']] for i, row in pairs_b.iterrows()]

                with warnings.catch_warnings():
                    warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)
                    # create lines between node pairs
                    for l in pairs_b:
                        # get coordinates and create LineString
                        node_from = nodes_int.loc[nodes_int[node_id] == l[0], 'geometry']
                        node_to = nodes_int.loc[nodes_int[node_id] == l[1], 'geometry']
                        line = LineString([[node_from.x, node_from.y], [node_to.x, node_to.y]])
                        # add to Lines GDF
                        lines.loc[len(lines) + 1] = [id_st, l[0], l[1], 999, b + '0', 0, 50, 0, line, np.nan]
                        id_st += 1
                self.dict_borders.update({b: len(pairs_b)})
            else:
                print('No border connection for {}'.format(b))
                self.dict_borders.update({b: 0})

        # concat border crossings and rest of the network
        self.network_int = pd.concat([self.network, lines])
        # ensure that IDs are integer
        self.network_int[roads_id] = self.network_int[roads_id].astype(int)
        print("Finished connecting border roads at {}".format(datetime.now()))



# =========================================================
# AttractionFactors.py
# @author Nina Thomsen
# @date 11.01.2023
# @copyright Institut fuer Verkehrsforschung,
#            Deutsches Zentrum fuer Luft- und Raumfahrt
# @brief generate attraction factors per TAZ and country attributes
# =========================================================

import warnings

import geopandas as gpd
import numpy as np
import osmnx as ox
import pandas as pd
from shapely.errors import ShapelyDeprecationWarning
from tqdm import tqdm


class AttractionIndex:
    """
    Calculate population and industrial areas per TAZ, transform to attraction index
    """

    def __init__(self, taz, taz_geo="geometry", taz_id="nuts_id"):
        """

        :param taz: GeoDataFrame with TAZ
        :param taz_geo: column name of geometry of taz
        :param taz_id: column name of id column in taz
        :type taz: gpd.GeoDataFrame
        :type taz_geo: str
        :type taz_id: str
        """
        self.taz = taz
        self.taz_geo = taz_geo
        self.taz_id = taz_id
        self.industry_ = gpd.GeoDataFrame()

    def population_from_point(self, pop_nodes, pop_val="VALUE"):
        """
        Aggregate population per TAZ based on point layer with population density
        
        --> result: self.taz as GeoDataFrame is updated with total population per cell

        :param pop_nodes: GeoDataFrame with population density (points)
        :param pop_val: name of column with population density in pop_nodes
        :return: None
        :type pop_nodes: gpd.GeoDataFrame
        :type pop_val: str
        """
        # overlay taz and population
        taz_pop = gpd.overlay(pop_nodes, self.taz)

        # aggregate pop per taz
        taz_pop_g = taz_pop.groupby(self.taz_id)[pop_val].sum().reset_index()
        taz_pop_g.rename(columns={pop_val: 'population'}, inplace=True)
        self.taz = self.taz.merge(taz_pop_g, how='left', on=self.taz_id)

    def industry_attributes_from_osm(self, return_taz=False):
        """
        Extract the number and aggregated area of polygons with landuse=industrial in OSM per TAZ

        Creates new columns in self.taz:
                - ind_area_count  | number of industrial sites
                - ind_area_sum    | aggregated area of industrial sites

        TAZ with new industry column can be returned by setting return_taz=True

        :param return_taz: if True, return a GeoDataFrame
        :type return_taz: bool
        :return: TAZ with industrial site count (ind_area_count) and area (ind_area_sum)
        :rtype: gpd.GeoDataFrame
        """
        taz_i = self.taz.copy()
        taz_i[['ind_area_count', 'ind_area_sum']] = 0
        # iterate over taz
        for i, t in tqdm(self.taz.iterrows()):
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)
                warnings.filterwarnings("ignore", category=FutureWarning, append=True)
                warnings.filterwarnings("ignore", category=UserWarning, append=True)
                industry_t = ox.geometries.geometries_from_polygon(t['geometry'], tags={"landuse": "industrial"})
            # only Polygon / Multipolygon geometries
            if len(industry_t.geom_type.unique()) > 1:
                industry_t = industry_t[industry_t.geom_type.isin(["MultiPolygon", "Polygon"])].copy()
            # get number of industrial areas
            taz_i.loc[i, 'ind_area_count'] = len(industry_t)
            # get total area
            industry_t.to_crs(epsg=3035, inplace=True)
            # industry_t['area'] = industry_t.area
            taz_i.loc[i, 'ind_area_sum'] = industry_t.area.sum()
        # set self.taz to taz_i with industrial area information
        self.taz = taz_i.copy()

        # return taz_i
        if return_taz:
            return taz_i

    def industry_attributes_from_gdf(self, industry_gdf, return_taz=False):
        """
        Aggregate total area and count of industrial sites per TAZ based on a GeoDataFrame with industrial areas
        
        --> result: self.taz as GeoDataFrame is updated with industrial area data per cell

         Creates new columns in self.taz:
                - ind_area_count  | number of industrial sites
                - ind_area_sum    | aggregated area of industrial sites

        :param industry_gdf: GeoDataFrame with industry areas (Polygons)
        :type industry_gdf: gpd.GeoDataFrame
        :param return_taz: if True, return a GeoDataFrame
        :type return_taz: bool
        :return: TAZ with industrial site count (ind_area_count) and area (ind_area_sum)
        :rtype: gpd.GeoDataFrame
        """

        if type(industry_gdf) == gpd.GeoDataFrame:
            industry = industry_gdf
            industry['id'] = list(range(len(industry)))
        else:
            raise ValueError('Wrong input type for industry_gdf {}'.format(str(type(industry_gdf))))

        # get area in km2 per polygon
        industry.to_crs(epsg=3035, inplace=True)
        industry['area_ind'] = industry.area / 1000**2
        industry.to_crs(epsg=4326, inplace=True)
        industry = industry[['id', 'area_ind', 'geometry']]
        industry.rename(columns={'id': 'id_ind'}, inplace=True)
        # aggregate industrial area per TAZ, number of industrial areas per TAZ
        taz_industry = gpd.overlay(industry, self.taz)
        taz_ind_area = taz_industry.groupby(self.taz_id)['area_ind'].sum().reset_index()
        taz_ind_area.rename(columns={'area_ind': 'ind_area_sum'}, inplace=True)
        taz_ind_count = taz_industry.groupby(self.taz_id)['id_ind'].aggregate('count').reset_index()
        taz_ind_count.rename(columns={'id_ind': 'ind_area_count'}, inplace=True)
        taz_industry = pd.merge(taz_ind_area, taz_ind_count, on=self.taz_id)
        self.taz = self.taz.merge(taz_industry, how='left', on=self.taz_id)

        # return taz
        if return_taz:
            return self.taz

    def attraction_index(self, scope=None, taz_cn='cntr_code', alpha=1.):
        """
        Create attraction index with Cobb-Douglas transformation
        :param scope: either look at all TAZ (None) or single country (str of ISO-code)
        :param taz_cn: column name for country identifier in self.taz
        :param alpha: alpha parameter for Cobb Douglas formula
        :type scope: str
        :type taz_cn: str
        :type alpha: float
        :return: GeoDataFrame with taz in scope and attraction index for scope
        """
        # set scope
        if scope is None:
            taz_scope = self.taz.copy()
            index_type = "index_int"
        else:
            taz_scope = self.taz[self.taz[taz_cn] == scope].copy()
            index_type = "index_nat"

        # columns of final geo data frame
        cols_taz = list(self.taz.columns)
        cols_taz.append(index_type)

        # normalize pop, industrial area to all taz
        taz_scope['pop_n'] = taz_scope['population'] / np.nanmean(taz_scope['population'])
        taz_scope['ind_a_n'] = taz_scope['ind_area_sum'] / np.nanmean(taz_scope['ind_area_sum'])
        taz_scope['ind_c_n'] = taz_scope['ind_area_count'] / np.nanmean(taz_scope['ind_area_count'])

        taz_scope[index_type] = alpha * (taz_scope['pop_n']**0.5 * taz_scope['ind_a_n']**0.25 * taz_scope['ind_c_n']**0.25)

        return taz_scope[cols_taz]


class BorderCrossingAtts:
    """
    Calculate country attributes like number of border crossing streets, neighboring countries that define the character of border-crossing road traffic
    """

    def __init__(self, taz, taz_cn="cntr_code", taz_geo="geometry"):
        """

        :param taz: GeoDataFrame with TAZ
        :param taz_cn: column name of country in TAZ
        :param taz_geo: column name of geometry in TAZ
        :type taz: gpd.GeoDataFrame
        :type taz_cn: str
        :type taz_geo: str
        """
        self.taz = taz
        self.countries = taz[taz_cn].unique()
        self.taz_cn = taz_cn
        self.taz_geo = taz_geo

        # get countries and init borders
        self.country_layer = taz.dissolve(by=taz_cn, aggfunc="sum").reset_index()
        self.country_layer = self.country_layer[[taz_cn, taz_geo]]

        self.border_layer = gpd.GeoDataFrame()

    def get_borderbuffer(self, buffer=5000):
        """
        Create GeoDataFrame with borders, using a defined buffer around these borders (i.e. with a 5000m buffer, there will be Polygons along borders with a width of 5000m)

        :param buffer: total buffer width in m, default 5000m
        :type buffer: float
        :return: GeoDataFrame with buffer polygons around borders
        """
        border = self.country_layer.copy()
        # set buffer around polygon borders (width of border polygons)
        border.to_crs(epsg=3035, inplace=True)
        border[self.taz_geo] = border[self.taz_geo].buffer(buffer/2)
        # find borders for each country
        gdf_borderbuffer = gpd.GeoDataFrame()

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=ShapelyDeprecationWarning)
            warnings.filterwarnings("ignore", category=FutureWarning, append=True)
            warnings.filterwarnings("ignore", category=UserWarning, append=True)

            for country in self.countries:

                country_ = border[border[self.taz_cn] == country]
                country_ = country_[self.taz_geo].buffer(1)

                for index, row in border.iterrows():

                    if row[self.taz_cn] != country:
                        intersec = country_.intersection(row[self.taz_geo].buffer(1))
                        if not intersec.values.is_empty[0]:
                            new_border = gpd.GeoDataFrame(geometry=intersec)
                            new_border[['country1', 'country2']] = [country, row[self.taz_cn]]
                            gdf_borderbuffer = pd.concat([gdf_borderbuffer, new_border])

            # merge borders create and set geometry
            borders = pd.merge(gdf_borderbuffer, gdf_borderbuffer, left_on=['country1', 'country2'],
                               right_on=['country2', 'country1'])
            borders['geometry'] = [r['geometry_x'].union(r['geometry_y']) for i, r in borders.iterrows()]
            borders['border'] = borders['country1_x'] + borders['country1_y']
            borders = borders.set_geometry('geometry')
            borders.crs = 3035
            # remove duplicates
            borders['abc'] = ["".join(sorted(row['border'])) for i, row in borders.iterrows()]
            borders.drop_duplicates(subset="abc", inplace=True)
            # finalize
            borders = borders[['border', 'country1_x', 'country1_y', 'geometry']]
            borders.rename(columns={'country1_x': 'country1', 'country1_y': 'country2'}, inplace=True)
            borders.to_crs(epsg=4326, inplace=True)
        
        return borders

    def shared_borders(self, inplace=False):
        """
        Determine the share of land borders of total country border. Islands would have a share of 0, while countries without a cost would land at 1.
        
        --> result: border shares are merged to self.country_layer

        :param inplace: return DataFrame; if True, DataFrame is not returned
        :type inplace: bool
        :return: pd.DataFrame with border length and share per country / if inplace=True return None
        """
        self.border_layer = self.get_borderbuffer(buffer=5000)
        border_shares = pd.DataFrame()
        crs_taz = self.country_layer.crs

        for country in self.countries:
            ctr_poly = self.country_layer[self.country_layer[self.taz_cn] == country]
            ctr_border_buffer = self.border_layer[(self.border_layer['country1'] == country) | (self.border_layer['country2'] == country)]

            ctr_border_line = gpd.GeoDataFrame([1], geometry=[ctr_poly[self.taz_geo].values[0].boundary])
            ctr_border_line.crs = crs_taz

            ctr_border_clipped = gpd.clip(ctr_border_line, ctr_border_buffer)
            ctr_border_clipped.crs = crs_taz

            # calculate length of total and shared border
            ctr_border_line.to_crs(epsg=3035, inplace=True)
            ctr_border_clipped.to_crs(epsg=3035, inplace=True)
            border_length = ctr_border_line.length.values[0] / 1000
            if len(ctr_border_clipped) > 0:
                border_length_shared = ctr_border_clipped.length.values[0] / 1000
            else:
                border_length_shared = 0
            ctr_border_share = (border_length_shared / border_length)

            border_shares = pd.concat([border_shares,
                                       pd.DataFrame({'country': country, 'border_share': ctr_border_share,
                                                     'border_length': border_length, 'border_length_shared': border_length_shared},
                                                    index=[0])])
        border_shares.reset_index(inplace=True)

        # merge to countries
        self.country_layer = pd.merge(self.country_layer, border_shares, how='left', left_on=self.taz_cn, right_on='country')
        self.country_layer.drop(['country'], axis=1, inplace=True)

        if not inplace:
            return border_shares

    def border_streets(self, net, net_type="type", type_filter=None, inplace=None):
        """
        Count the number of border crossing streets per country
        
        --> result: number of border crossings is merged to self.country_layer

        :param net: GeoDataFrame with international road network
        :param net_type: name of column specifiying road type in net
        :param type_filter: list of road types to filter for border crossing roads; default None, meaning types [1,2]
        :param inplace: return DataFrame; if True, DataFrame is not returned
        :type net: gpd.GeoDataFrame
        :type net_type: str
        :type type_filter: list
        :type inplace: bool
        :return: pd.DataFrame with number of border crossing streets per country / if inplace=True return None
        """

        if type_filter is None:
            type_filter = [1, 2]
        # filter road types to count for border crossings
        net = net[net[net_type].isin(type_filter)]
        # get streets within border layer
        border_2m = self.get_borderbuffer(buffer=2)
        borderbuffer_streets = gpd.overlay(net, border_2m, how="union", keep_geom_type=True)
        # aggregate streets per country and border
        country1_grp = borderbuffer_streets.groupby('country1')['border'].aggregate('count').reset_index()
        country2_grp = borderbuffer_streets.groupby('country2')['border'].aggregate('count').reset_index()
        country1_grp = country1_grp.rename(columns={'country1': 'country'})
        country2_grp = country2_grp.rename(columns={'country2': 'country'})
        country_grp = pd.concat([country1_grp, country2_grp])
        country_grp = country_grp.groupby('country')['border'].sum().reset_index()
        country_grp.rename(columns={'border': 'border_crossings_count'}, inplace=True)

        # merge to countries
        self.country_layer = pd.merge(self.country_layer, country_grp, how='left', left_on=self.taz_cn, right_on='country')
        self.country_layer.drop(['country'], axis=1, inplace=True)

        if not inplace:
            return country_grp

    def count_neighbors(self, inplace=False):
        """
        Determine the number of direct neighbor countries (shared border)
        
        --> result: number of neighbors is merged to self.country_layer

        :param inplace: return DataFrame; if True, DataFrame is not returned
        :type inplace: bool
        :return: pd.DataFrame with number of neighbor countries per country / if inplace=True return None
        """
        dict_ = {}
        for c in self.countries:
            tab = self.border_layer[(self.border_layer['country1'] == c) | (self.border_layer['country2'] == c)]
            adj = len(tab['border'].unique())
            dict_.update({c: {'country': c, 'neighbors': adj}})
        neighbors = pd.DataFrame.from_dict(dict_, orient='index')
        # merge to countries
        self.country_layer = pd.merge(self.country_layer, neighbors, how='left', left_on=self.taz_cn, right_on='country')
        self.country_layer.drop(['country'], axis=1, inplace=True)
        if not inplace:
            return neighbors

    def pop_area(self, pop_values=None, pop_cn="country", taz_pop="population", inplace=False):
        """
        Determine total population and area im km² per country. Population can be determined by aggregating population per taz or with and external input DataFrame
        
        --> result: population and area are merged to self.country_layer

        :param pop_values: pd.DataFrame with population per country; if None, population will be aggregated from taz attributes, default None
        :param pop_cn: name of column with country name in pop_values
        :param taz_pop: name of column with population in self.taz
        :param inplace: return GeoDataFrame with border shares; if True, DataFrame is not returned
        :type pop_values: pd.DataFrame
        :type pop_cn: str
        :type taz_pop: str
        :type inplace: bool
        :return: GeoDataFrame with countries and their population and area / if inplace=True return None
        """
        if pop_values is None:
            # aggregate population per country from taz
            pop_agg = self.taz.groupby(self.taz_cn)[taz_pop].sum().reset_index()
            self.country_layer = pd.merge(self.country_layer, pop_agg, how='left', on=self.taz_cn)
        else:
            self.country_layer = pd.merge(self.country_layer, pop_values, how='left', left_on=self.taz_cn, right_on=pop_cn)

        # calculate area in km2
        self.country_layer.to_crs(epsg=3035, inplace=True)
        self.country_layer['area'] = self.country_layer[self.taz_geo].area / 1000**2
        self.country_layer.to_crs(epsg=4326, inplace=True)

        if not inplace:
            return self.country_layer



# =========================================================
# DistributeTraffic.py
# @author Nina Thomsen
# @date 02.02.2023
# @copyright Institut fuer Verkehrsforschung,
#            Deutsches Zentrum fuer Luft- und Raumfahrt
# @brief Distribute traffic between TAZ and create OD Matrices;
#        Assign inner cell traffic to network links
# =========================================================

import itertools

import networkx as nx
import numpy as np
import osmnx as ox
import pandas as pd


def pt_shares_int(a, b):
    """
    Determine shares of international transport for personal transport for a country based on area and share of land border

    :param a: area of country
    :param b: share of land border of country
    :return: dict with international transport shares
    :type a: float
    :type b: float
    :rtype: dict
    """
    # transform area
    a = (np.log((a-(-15000))/6.59)/0.437)-22.69

    x1 = (0.3 / (1 + ((0.3 / 0.095 - 1) * np.exp(a * 0.3 * -(-1.05))))) * b
    x2 = max(0.02, (0.35 + -0.04 * a) * b**0.5)
    x3 = min(0.95, max(0.1, (0.3 + 0.05 * a)))

    return {'pt_share_foreign': x1, 'pt_share_transit': x2, 'pt_ratio_int': x3}


def get_trip_matrix(mx_grav, taz, goal_col, taz_id='id'):
    """
    Get trip matrix based on total trips per taz and gravity model matrix

    :param mx_grav: matrix from gravity model
    :param taz: gpd.GeoDataFrame or pd.DataFrame with TAZ; needs id column to identify coordinate in mx_grav
    :param goal_col: name of column in taz with total trips per taz
    :return: matrix with trips between taz
    :type mx_grav: np.array
    :type taz: gpd.GeoDataFrame or pd.DataFrame
    :type goal_col: str
    :rtype: np.array
    """
    # choice probabilities and trips
    mx_trips = np.zeros(mx_grav.shape)
    for t in taz[taz_id]:
        goal = float(taz.loc[taz[taz_id] == t, goal_col])
        # get probabilities for trips starting at t
        prob_a = mx_grav[t, :] / mx_grav[t, :].sum()
        prob_b = mx_grav[:, t] / mx_grav[:, t].sum()
        mx_trips[t, :] += prob_a * (goal / 2)
        mx_trips[:, t] += prob_b * (goal / 2)
    # fill NA
    mx_trips[np.isnan(mx_trips)] = 0
    return mx_trips


def assignment_single(net_g, net, o, d, weight, trips, from_='from_node', to_='to_node'):
    """
    Performs a shortest path network assignment between two nodes on a MultiDiGraph based on a specified weight and returns a GeoDataFrame with the resulting network load
    :param net_g: MultiDiGraph with network
    :param net: gpd.GeoDataFrame with network
    :param o: ID of origin node
    :param d: ID of destination node
    :param weight: name of weight attribute; is an attribute of links in net and net_g, e.g. travel time or length
    :param trips: number of vehicle trips to distribute between o and d
    :param from_: name of from node attribute in net, defaults to 'from_node'
    :param to_: name of to node attribute in net, defaults to 'to_node'
    :return: GeoDataFrame of net with trips on the links, on the path between o and d

    :type net_g: MultiDiGraph
    :type net: gpd.GeoDataFrame
    :type o: int
    :type d: int
    :type weight: str
    :type trips: float
    :type from_: str
    :type to_: str
    :rtype: gpd.GeoDataFrame
    """
    path_nodes = ox.distance.shortest_path(net_g, o, d, weight=weight)  # tt or length
    nodes_tuples = [(path_nodes[x], path_nodes[x + 1]) for x in range(len(path_nodes) - 1)]
    nodes_tuples = pd.DataFrame(nodes_tuples, columns=[from_, to_])
    nodes_tuples['__tmp'] = trips
    # add trips to network links between nodes tuples
    net_tmp = net.merge(nodes_tuples, on=[from_, to_], how='left')
    net_tmp.loc[net_tmp['__tmp'].isna(), '__tmp'] = 0
    return net_tmp


def assignment_multiple(net_g, net, o, d, k, weight, trips, from_='from_node', to_='to_node', id_='link_id'):
    """
    Performs a shortest path assignment for the k shortest paths between two nodes on a MultiDiGraph based on a specified weight and returns a GeoDataFrame with the resulting network load
    Paths are weighted based on the total weight per path
    :param net_g: MultiDiGraph with network
    :param net: gpd.GeoDataFrame with network
    :param o: ID of origin node
    :param d: ID of destination node
    :param k: int, number of paths to find
    :param weight: name of weight attribute; is an attribute of links in net and net_g, e.g. travel time or length
    :param trips: number of vehicle trips to distribute between o and d
    :param from_: name of from node attribute in net, defaults to 'from_node'
    :param to_: name of to node attribute in net, defaults to 'to_node'
    :param id_: name of id attribute in net, defaults to 'link_id'
    :return: GeoDataFrame of net with trips on the links, on the path between o and d

    :type net_g: MultiDiGraph
    :type net: gpd.GeoDataFrame
    :type o: int
    :type d: int
    :type k: int
    :type weight: str
    :type trips: float
    :type from_: str
    :type to_: str
    :type id_: str
    :rtype: gpd.GeoDataFrame
    """
    # transform to digraph for nx.shortest_simple_path function
    dg = nx.DiGraph(net_g)
    k_paths = nx.shortest_simple_paths(dg, o, d, weight)
    net_tmp = net[[from_, to_, id_, weight]].copy()
    net_tmp['__weight_all'] = 0
    weights = []
    for path_nodes in itertools.islice(k_paths, 0, k):
        nodes_tuples = [(path_nodes[x], path_nodes[x + 1]) for x in range(len(path_nodes) - 1)]
        nodes_tuples = pd.DataFrame(nodes_tuples, columns=[from_, to_])
        # merge net_tmp, aggregate weight (tt or length) for current path
        nodes_tuples = nodes_tuples.merge(net[[from_, to_, weight]], on=[from_, to_], how='left')
        weight_ = 1 / nodes_tuples[weight].sum()
        nodes_tuples['__weight'] = weight_
        weights.append(weight_)
        # merge weight to net_tmp and add to weight_all
        net_tmp = net_tmp.merge(nodes_tuples[[from_, to_, '__weight']], on=[from_, to_], how='left')
        net_tmp.loc[net_tmp['__weight'].isna(), '__weight'] = 0
        net_tmp['__weight_all'] += net_tmp['__weight']
        net_tmp.drop(columns=['__weight'], inplace=True)
    k_paths.close()
    net_tmp['__weight_all'] /= sum(weights)
    net_tmp['__tmp'] = net_tmp['__weight_all'] * trips
    return net_tmp


def create_network_graph(net_, nodes_, node_id='node_id', node_geo='geometry', net_from='from_node', net_to='to_node', net_id='link_id'):
    """
    Transform a GeoDataFrame of a network and the corresponding nodes to a MultiDiGraph
    Needs columns for node id in nodes, from node, to node and link id in network
    :param net_: GeoDataFrame with network
    :param nodes_: GeoDataFrame with nodes
    :param node_id: name of id column in nodes_, defaults to 'node_id'
    :param node_geo: name of geometry column in nodes_, defaults to 'geometry'
    :param net_from: name of from node column in net_, defaults to 'from_node'
    :param net_to: name of to node column in net_, defaults to 'to_node'
    :param net_id: name of id column in net_, defaults to 'link_id'
    :return: MultiDiGraph of net_

    :type net_: gpd.GeoDataFrame
    :type nodes_: gpd.GeoDataFrame
    :type node_id: str
    :type node_geo: str
    :type net_from: str
    :type net_to: str
    :type net_id: str
    :rtype: MultiDiGraph
    """
    nodes_.set_index(node_id, inplace=True)
    nodes_['x'] = nodes_[node_geo].x
    nodes_['y'] = nodes_[node_geo].y
    net_.set_index([net_from, net_to, net_id], inplace=True)

    net_g = ox.utils_graph.graph_from_gdfs(nodes_, net_)
    net_g = ox.utils_graph.get_largest_component(net_g)

    return net_g


class TargetValues:
    """
    Determine shares of international and national transport for a country and calculate corresponding total trips and vehicle kilometers travelled (VKT)
    """

    def __init__(self, country_layer, cn_col='cntr_code'):
        """

        :param country_layer: countries with target values for raod transport volumes
        :param cn_col: name of column with country code in country layer, default 'cntr_code'
        :type country_layer: pd.DataFrame or gpd.GeoDataFrame
        :type cn_col: str
        """
        self.country_layer = country_layer
        self.cn_col = cn_col

    def targets_personal_transport(self, cn, target_col='car_pkm', unit=1.e+6):
        """
        Calculate VKT for the segments short and long distance (national) and international personal transport for a country

        :param cn: country code
        :param target_col: name of column in self.country_layer with the total transport volume in person kilometers travelled, default 'car_pkm'
        :param unit: scaling factor for total transport volume, default 1.e+6
        :return: vkt per segment

        :type cn: str
        :type target_col: str
        :type unit: float
        :rtype: dict
        """
        target = float(self.country_layer.loc[self.country_layer[self.cn_col]==cn, target_col])
        shares_int = pt_shares_int(a=float(self.country_layer.loc[self.country_layer[self.cn_col]==cn, 'area']),
                                   b=float(self.country_layer.loc[self.country_layer[self.cn_col]==cn, 'border_share']))

        f_in_c = target * shares_int['pt_share_foreign']
        int1 = target * shares_int['pt_share_foreign'] * shares_int['pt_ratio_int']
        inner = target - f_in_c - int1

        short_ = 0.65 * (inner + int1)
        long_ = 0.35 * (inner + int1) - int1

        trans_ = shares_int['pt_share_transit'] * f_in_c
        int_ = (f_in_c - trans_) + int1

        return {"short": unit * short_, "long": unit * long_, "int": unit * int_, "trans": unit * trans_}

    def targets_freight_transport(self, cn, target_col='freight_tkm', segments=None, shares_tkm=None, loads_t=None, seg_split=None, unit=1.e+6):
        """
        Calculate VKT and trips for the segments short and long distance (national) and international freight transport for a country

        !! Long-distance and international freight transport only cover 1 segment (heavy weight transport)
        !! Short-distance freight transport can cover multiple segments, inlcuding the long-distance segment (heavy weight transport)
        !! The distribution of tkm among segments is given with the shares_tkm parameter, short-distance tkm are then further spread among these segments

        :param seg_split: freight transport segment where there is traffic in all categories (short, long, international)
        :param cn: country code
        :param target_col: name of column in self.country_layer with the total transport volume in tkm
        :param segments: names of freight transport segments; default None refers to ['lcv', 'mft', 'hft']
        :param shares_tkm: shares of different freight transport segments of total tkm; default None refers to shares {'lcv': 0.08, 'mft': 0.09, 'hft': 0.83}
        :param loads_t: average load in t per freight transport segment; default None refers to load {'lcv': 0.5, 'mft': 3, 'hft': 10}
        :param unit: scaling factor for total transport volume, default 1.e+6
        :return: vkt and trips per segment

        :type seg_split: str
        :type cn: str
        :type target_col: str
        :type segments: list
        :type shares_tkm: dict
        :type loads_t: dict
        :type unit: float
        :rtype: dict
        """

        if shares_tkm is None:
            shares_tkm = {'lcv': 0.08, 'mft': 0.09, 'hft': 0.83}
        if loads_t is None:
            loads_t = {'lcv': 0.5, 'mft': 3, 'hft': 10}
        if seg_split is None:
            seg_split = 'hft'
        if segments is None:
            segments = shares_tkm.keys()

        # check if segments, loads and shares match
        if (sorted(segments) == sorted(shares_tkm.keys())) & (sorted(segments) == sorted(loads_t.keys())) & (seg_split in segments):

            target = float(self.country_layer.loc[self.country_layer[self.cn_col] == cn, target_col])

            cat_dict = {0: {'int_imex': 0.03, 'int_transit': 0.00, 'nat': 0.97},
                        1: {'int_imex': 0.15, 'int_transit': 0.02, 'nat': 0.83},
                        2: {'int_imex': 0.15, 'int_transit': 0.05, 'nat': 0.80},
                        3: {'int_imex': 0.20, 'int_transit': 0.10, 'nat': 0.70},
                        4: {'int_imex': 0.00, 'int_transit': 0.00, 'nat': 1.00}}

            border_share = float(self.country_layer.loc[self.country_layer[self.cn_col] == cn, 'border_share'])
            neighbors = float(self.country_layer.loc[self.country_layer[self.cn_col] == cn, 'neighbors'])
            border_crossings = float(self.country_layer.loc[self.country_layer[self.cn_col] == cn, 'border_crossings_count'])

            # find category for country
            cat_country = 2
            if (border_share < 0.1) & (neighbors <= 1):
                cat_country = 0
            elif neighbors <= 5:
                if (border_crossings < 4) | (border_share < 0.5):
                    cat_country = 1
                else:
                    cat_country = 2
            elif neighbors >= 5:
                if border_share < 0.75:
                    cat_country = 2
                else:
                    cat_country = 3
            elif border_share == 0:
                cat_country = 4

            # distances
            area = float(self.country_layer.loc[self.country_layer[self.cn_col] == cn, 'area'])
            dis_long = 0.365 * area**(1/2) + 61.488
            dis_short = 1.183 * area**(1/4) + 22.361
            dis_transit = 0.928 * area**(1/2) + 38.998

            # transform input tkm to t:
            share_national = cat_dict[cat_country]['nat']
            share_international = cat_dict[cat_country]['int_imex']
            share_transit = cat_dict[cat_country]['int_transit']

            t_tot = target / ((dis_long * 0.2 + dis_short * 0.8) * share_national + dis_long * share_international + dis_transit * share_transit)
            t_nat = t_tot * share_national
            t_int = t_tot * share_international
            t_tra = t_tot * share_transit

            # tkm
            tkm_long = t_nat * 0.2 * dis_long
            tkm_short = t_nat * 0.8 * dis_short
            tkm_imex = t_int * dis_long
            tkm_transit = t_tra * dis_transit
            tkm = sum([tkm_imex, tkm_short, tkm_long, tkm_transit])

            tkm_dict = {'short': tkm_short, 'long': tkm_long}

            # transform tkm to vkm by using average tkm shares and average load
            # short distance transport: split into vehicle segments
            vkm_short_segments = {}
            for s in segments:
                # total
                tkm_s = shares_tkm[s] * tkm
                # short distance only segments
                if s != seg_split:
                    vkm_ts = tkm_s / loads_t[s]
                else:
                    # determine short distance tkm
                    tkm_s_short = tkm_s - (tkm - tkm_short)
                    vkm_ts = tkm_s_short / loads_t[s]
                vkm_short_segments.update({s: vkm_ts * unit})
            vkm_short = sum([vkm_short_segments[s] for s in segments])

            # vkm long-distance and international
            vkm_long = tkm_long / loads_t[seg_split] * unit
            vkm_imex = tkm_imex / loads_t[seg_split] * unit
            vkm_transit = tkm_transit / loads_t[seg_split] * unit

            # result dictionary
            result = {'short': vkm_short, 'trips_short': vkm_short / dis_short,
                      'long': vkm_long, 'trips_long': vkm_long / dis_long,
                      'int': vkm_imex, 'trips_int': vkm_imex / dis_long,
                      'transit': vkm_transit}

            result.update({'short_segments_vkm': vkm_short_segments})

            return result
        else:
            raise KeyError('Segments and shares / loads do not match!')


class GravityModel:
    """
    Create OD trip matrices for personal and freight transport using a gravity model
    """

    def __init__(self, cost_matrix, taz, taz_cn="cntr_code"):
        """

        :param matrix: np.array with shape [len(taz),len(taz),2], cost matrix with travel times [:,:,0] and distances [:,:,1]
        :param taz: gpd.GeoDataFrame or pd.DataFrame with TAZ
        :param taz_cn: name of column with country code in TAZ

        :type matrix: np.array
        :type taz: gpd.GeoDataFrame or pd.DataFrame
        :type taz_cn: str
        """
        self.matrix = cost_matrix
        self.taz = taz
        self.taz_cn = taz_cn

    def get_country_model(self, cn):
        """
        Get TAZ and cost matrix for country
        :param cn: country code
        :return: TAZ for country, cost matrix for country
        :type cn: str
        :rtype: pd.DataFrame or gpd.GeoDataFrame, np.array
        """
        taz_c = self.taz[self.taz[self.taz_cn] == cn].copy()
        taz_c_ids = taz_c['id']
        # matrix from country zones
        mtx_cn = self.matrix[taz_c_ids, :, :]
        # matrix to country zones
        mtx_cn = mtx_cn[:, taz_c_ids, :]
        # set new taz id for taz_c
        new_id = np.c_[taz_c_ids, range(len(taz_c_ids))]
        taz_c['id'] = new_id[:, 1]
        return taz_c, mtx_cn

    def matrix_international(self):
        """
        Adjust cost matrix for international assignment (no inner country trips) by setting costs for inner country relations to 9.e+12

        :return: international cost matrix with travel times and distances
        :rtype: np.array
        """
        countries = self.taz[self.taz_cn].unique()
        mx_int = self.matrix.copy()
        mx_fil = np.zeros(self.matrix.shape)
        for cn in countries:
            # find all IDs for cn and set values between them to 9.e+12
            taz_c_ids = self.taz.loc[self.taz[self.taz_cn]==cn, 'id'].tolist()
            mx_fil[taz_c_ids, :, :] += 1
            mx_fil[:, taz_c_ids, :] += 1
            mx_fil[mx_fil == 1] = 0
        mx_int[mx_fil > 1] = 9.e+12
        return mx_int

    def trip_distribution_pt(self, target, cn=None, taz_pop='population', alpha=0.5, gamma=-2.75, unit_dis=1000, unit_tt = 60, mob_rate=36., occ_rate=1.3):
        """
        Distribute personal transport using a gravity model and an input for total VKT
        Gravity model parameters are given as defaults and were estimated using the German NHTS MiD 2017

        :param target: total VKT
        :param cn: country code; default None means all countries in self.taz are included
        :param taz_pop: name of column with population in taz, default 'population'
        :param alpha: alpha parameter in gravity model, default 0.5
        :param gamma: gamma parameter in gravity model, default -2.75
        :param unit_dis: factor to transform unit in distance matrix to km, default 1000 (suggesting distance is in m)
        :param unit_tt: factor to transform unit in travel time matrix to min, default 60 (suggesting distance is in s)
        :param mob_rate: mobility rate of inhabitants, default 36
        :param occ_rate: average vehicle occupancy
        :return: np.array with OD trip matrix

        :type target: float
        :type cn: str
        :type taz_pop: str
        :type alpha: float
        :type gamma: float
        :type unit_dis: float
        :type unit_tt: float
        :type mob_rate: float
        :type occ_rate: float
        :rtype: np.array
        """
        if cn is not None:
            taz, mtx = self.get_country_model(cn=cn)
        else:
            taz = self.taz
            mtx = self.matrix_international()
            #mtx[mtx==9.e+12] = 0

        # transform units in mtx to min, km
        mtx[:, :, 0] /= unit_tt
        mtx[:, :, 1] /= unit_dis

        # trip generation
        taz['pt_goal'] = taz[taz_pop] * mob_rate

        # trip distribution
        # gravity model: gravity values
        mx_grav = np.zeros((len(taz), len(taz)))
        for o in taz['id']:
            pop_o = float(taz.loc[taz['id'] == o, taz_pop])
            for d in taz['id']:
                pop_d = float(taz.loc[taz['id'] == d, taz_pop])
                mx_grav[o, d] = (pop_o*pop_d)**alpha * mtx[o, d, 0]**gamma
        # fill diagonal (no trips) and NaN
        np.fill_diagonal(mx_grav, 0)
        mx_grav[np.isnan(mx_grav)] = 0
        # choice probabilities and trips
        mx_trips = get_trip_matrix(mx_grav, taz, 'pt_goal', taz_id='id')
        # scaling to match target
        vkt = (mx_trips*mtx[:, :, 1]).sum()
        scale_fac = target / vkt
        print('Scaling factor personal transport for {}: {}'.format(cn, scale_fac))
        mx_trips *= scale_fac

        # transform person trips to vehicle trips
        mx_trips /= occ_rate

        return mx_trips

    def trip_distribution_ft(self, target_trips, target_vkt=None, cn=None, beta=0.00421, unit_tt=60, unit_dis=1000, trips_key=''):
        """
        Distribute freight transport using a gravity model and an input for total trips and VKT
        Gravity model parameters are given as defaults and were estimated using microscopic truck data for Germany
        :param target_trips: float or dict, total trips or if cn is None dict with total trips per country and segment in the form of {cn: {segment: target_trips}}
        :param target_vkt: optional: total VKT used for comparing VKT from OD matrix and goal, default None
        :param cn: country code; default None means all countries in self.taz are included
        :param beta: gamma parameter in gravity model, default 0.00421
        :param unit_dis: factor to transform unit in distance matrix to km, default 1000 (suggesting distance is in m)
        :param unit_tt: factor to transform unit in travel time matrix to min, default 60 (suggesting distance is in s)
        :param trips_key: key for segment if cn is None target_trips is dict
        :return: np.array with OD trip matrix

        :type target_trips: float or dict
        :type target_vkt: float
        :type cn: None or str
        :type beta: float
        :type unit_dis: float
        :type unit_tt: float
        :type trips_key: str
        :rtype: np.array
        """
        t_targ = type(target_trips)
        if cn is not None:
            taz, mtx = self.get_country_model(cn=cn)
            index_col = 'index_nat'
            if (t_targ != float) & (t_targ != int):
                raise ValueError('target_trips has to be float / int if cn is not None, is {}'.format(t_targ))
        else:
            taz = self.taz
            mtx = self.matrix_international()
            index_col = 'index_int'
            if t_targ != dict:
                raise ValueError('target_trips has to be dict if cn is None, is {}'.format(t_targ))

        # transform units in mtx to min, km
        mtx[:, :, 0] /= unit_tt
        mtx[:, :, 1] /= unit_dis

        # trip generation
        if cn is not None:
            taz['ft_goal'] = taz[index_col]/taz[index_col].sum() * target_trips
        else:
            # trip generation per country using target values
            taz['ft_goal'] = 0
            for c in taz[self.taz_cn].unique():
                if c in target_trips.keys():
                    taz.loc[taz[self.taz_cn]==c, 'ft_goal'] = taz.loc[taz[self.taz_cn]==c, index_col]/taz.loc[taz[self.taz_cn]==c, index_col].sum() * target_trips[c][trips_key]

        # trip distribution
        # gravity model: gravity values
        mx_grav = np.zeros((len(taz), len(taz)))
        for o in taz['id']:
            ind_o = float(taz.loc[taz['id'] == o, index_col])
            for d in taz['id']:
                ind_d = float(taz.loc[taz['id'] == d, index_col])
                mx_grav[o, d] = ind_o * ind_d * np.exp(-beta*mtx[o, d, 1])
        # fill diagonal (no trips) and NaN
        np.fill_diagonal(mx_grav, 0)
        mx_grav[np.isnan(mx_grav)] = 0

        # get trips
        mx_trips = get_trip_matrix(mx_grav, taz, 'ft_goal', taz_id='id')

        if target_vkt is not None:
            # determine relation to target
            vkt = (mx_trips * mtx[:, :, 1]).sum()
            scale_fac = target_vkt / vkt
            print('Freight transport gravity model result for {}: Relation to target vkm {}'.format(cn, scale_fac))

        return mx_trips


class IntraZonal:
    """
    Distribute intrazonal trips on road network
    """

    def __init__(self, taz, net, from_node='from_node', to_node='to_node', link_id='ultimo_id'):
        """

        :param taz: gpd.GeoDataFrame or pd.DataFrame with TAZ
        :param net: gpd.GeoDataFrame or pd.DataFrame with road network
        :param from_node: column name for from node in net, default 'from_node'
        :param to_node: column name for to node in net, default 'to_node'
        :param link_id: column name for unique id of the elements in net, default 'ultimo_id'
        :type taz: gpd.GeoDataFrame or pd.DataFrame
        :type net: gpd.GeoDataFrame or pd.DataFrame
        :type from_node: str
        :type to_node: str
        :type link_id: str
        """
        self.taz = taz
        self.net = net
        self.from_ = from_node
        self.to_ = to_node
        self.link_id = link_id

    def road_type_weighted_single(self, target, weights=None, veh_types=None, taz_id='nuts_id',
                                  index='population', sub_len='length_sub', net_type='type', occ_rate=1.):
        """
        Distribute total VKT per TAZ and assign loads to roads within this TAZ, weighted by road type

        :param target: target values for total VKT per vehicle type in the form of {veh_type: target_vkt}
        :param weights: weights of different road types; default None leads to pd.DataFrame({net_type: [0, 1, 2, 3], 'weight': [0.75, 1.5, 3.5, 3.5]})
        :param veh_types: names of vehicle types to be considered; default None leads to ['car']
        :param taz_id: name of column with taz id in self.taz and self.net, defaults to 'nuts_id'
        :param index: name of column with attraction index to be used for trip generation in self.taz, defaults to 'population'
        :param sub_len: name of column with length of subordinate network in self.taz, defaults to 'length'
        :param net_type: name of column with road type in self.net, defaults to 'type'
        :param occ_rate: average vehicle occupancy (applied for personal transport, i.e. if veh_type == 'car')
        :return: gpd.GeoDataFrame or pd.DataFrame for network with traffic loads, taz with subordinate network VKT
        :type target: dict
        :type weights: None or pd.DataFrame
        :type veh_types: list
        :type taz_id: str
        :type index: str
        :type sub_len: str
        :type net_type: str
        :type occ_rate: float
        :rtype: gpd.GeoDataFrame or pd.DataFrame
        """
        if weights is None:
            weights = pd.DataFrame({net_type: [0, 1, 2, 3], 'weight': [0.75, 1.5, 3.5, 3.5]})
        if veh_types is None:
            veh_types = ['car']

        # subordinate network length from TAZ
        sub = self.taz[[taz_id, sub_len]].copy()
        sub.rename(columns={sub_len: 'length'}, inplace=True)
        sub[net_type] = 0
        sub['weighted_length'] = sub['length']*float(weights.loc[weights[net_type]==0, 'weight'])
        # concat with network length per taz from net
        len_per_type = self.net.groupby([taz_id, net_type])['length'].sum().reset_index()
        len_per_type = pd.concat([len_per_type, sub[[taz_id, net_type, 'length']]])

        # weighted length per net type and taz
        net = self.net.merge(weights, on=net_type, how='left')
        net.loc[net['weight'].isna(), 'weight'] = 0
        net['weighted_length'] = net['length']*net['weight']

        w_len_per_type = net.groupby([taz_id, net_type])['weighted_length'].sum().reset_index()
        w_len_per_type = pd.concat([w_len_per_type, sub[[taz_id, net_type, 'weighted_length']]])

        net.drop(columns=['weight', 'weighted_length'], inplace=True)

        # weighted length per taz
        w_len_per_taz = w_len_per_type.groupby(taz_id)['weighted_length'].sum().reset_index()
        w_len_per_taz.rename(columns={'weighted_length':'sum_weighted_length'}, inplace=True)

        # merge len and w_len
        len_per_type = len_per_type.merge(w_len_per_type, on=[taz_id, net_type], how='left')

        # merge weighted length sum to taz
        taz_2 = self.taz.merge(w_len_per_taz, on=taz_id, how='left')

        del w_len_per_taz, w_len_per_type

        taz_result = self.taz.copy()
        for veh_type in veh_types:

            vkt_p_p = target[veh_type] / self.taz[index].sum()

            if veh_type == 'car':
                # apply occupancy rate
                vkt_p_p /= occ_rate

            # calculate vkt per taz
            taz_2['goal_taz_{}'.format(veh_type)] = taz_2[index] * vkt_p_p
            taz_2['goal_norm_{}'.format(veh_type)] = taz_2['goal_taz_{}'.format(veh_type)] / taz_2['sum_weighted_length']

            # merge to len_per_type
            len_per_type = len_per_type.merge(taz_2[[taz_id, 'goal_taz_{}'.format(veh_type), 'goal_norm_{}'.format(veh_type)]], on=taz_id, how='left')
            len_per_type['vkt'] = len_per_type['goal_norm_{}'.format(veh_type)] * len_per_type['weighted_length']
            len_per_type['trips'] = len_per_type['vkt'] / len_per_type['length']

            # add to network links and taz
            net = net.merge(len_per_type[[taz_id, net_type, 'trips']], on=[taz_id, net_type], how='left')
            net.loc[net['trips'].isna(), 'trips'] = 0
            net.rename(columns={'trips': '{}_short'.format(veh_type)}, inplace=True)

            sec = len_per_type[len_per_type[net_type] == 0]
            taz_result = taz_result.merge(sec[[taz_id, 'vkt']], on=taz_id, how='left')
            taz_result.rename(columns={'vkt': '{}_sub'.format(veh_type)}, inplace=True)

        return net, taz_result

    def road_type_weighted_multiple(self, target, matrix_dis, veh_types=None, weights=None, taz_id='nuts_id', taz_mx_id='id',
                                    index='index_nat', sub_len='length_sub', net_type='type', distance=55, cell_size=500,
                                    fac_cell=1.5, occ_rate=1.):
        """
        Distribute total VKT per TAZ and assign loads to roads within this TAZ and surrounding TAZ, weighted by road type

        :param target: target values for total VKT per vehicle type in the form of {veh_type: target_vkt}
        :param matrix_dis: distance matrix between TAZ (in km)
        :param veh_types: names of vehicle types to be considered; default None leads to ['lcv', 'mft', 'hft']
        :param weights: weights of different road types; default None leads to pd.DataFrame({net_type: [0, 1, 2, 3], 'weight': [0, 5, 1.5, 0.5]})
        :param taz_id: name of column with taz id in self.taz and self.net, defaults to 'nuts_id'
        :param taz_mx_id: name of column with taz id relating to position in matrix_dis in self.taz, defaults to 'id'
        :param index: name of column with attraction index to be used for trip generation in self.taz, defaults to 'index_nat'
        :param sub_len: name of column with length of subordinate network in self.taz, defaults to 'length'
        :param net_type: name of column with road type in self.net, defaults to 'type'
        :param distance: max distance between TAZ to be included in km, defaults to 55km
        :param cell_size: max cell size of TAZ to force inclusion of surrounding TAZ, defaults to 500km²
        :param fac_cell: factor to apply to main TAZ during distribution, defaults to 1.5
        :param occ_rate: average vehicle occupancy (applied for personal transport, i.e. if veh_type == 'car')
        :return: gpd.GeoDataFrame or pd.DataFrame for network with traffic loads, taz with subordinate network VKT
        :type target: dict
        :type matrix_dis: np.array
        :type veh_types: list
        :type weights: pd.DataFrame
        :type taz_id: str
        :type taz_mx_id: str
        :type index: str
        :type sub_len: str
        :type net_type: str
        :type distance: float
        :type cell_size: float
        :type fac_cell: float
        :type occ_rate: float
        :rtype: gpd.GeoDataFrame or pd.DataFrame
        """

        if weights is None:
            weights = pd.DataFrame({net_type: [0, 1, 2, 3], 'weight': [0, 5, 1.5, 0.5]})
        if veh_types is None:
            veh_types = ['lcv', 'mft', 'hft']

        net = self.net.copy()
        # target per taz
        taz_2 = self.taz[[taz_id, taz_mx_id, index, 'area', sub_len]].copy()
        for veh_type in veh_types:
            taz_2[veh_type] = taz_2[index] / taz_2[index].sum() * target[veh_type]
            if veh_type == 'car':
                # apply occupancy rate
                taz_2[veh_type] /= occ_rate
            net['{}_short'.format(veh_type)] = 0
            taz_2['{}_sub'.format(veh_type)] = 0

        # weighted length per net type and taz
        net = net.merge(weights, on=net_type, how='left')
        net.loc[net['weight'].isna(), 'weight'] = 0
        net['weighted_length'] = net['length'] * net['weight']
        taz_2['weighted_sub'] = taz_2[sub_len]*float(weights.loc[weights[net_type]==0, 'weight'])

        # iterate over taz
        for t in taz_2[taz_id]:
            # adjust link weight for roads within taz
            net['weighted_length2'] = net['weighted_length']
            net['weight2'] = net['weight']
            taz_2['weighted_sub2'] = taz_2['weighted_sub']
            taz_2['weight2'] = float(weights.loc[weights[net_type]==0, 'weight'])
            net.loc[net[taz_id] == t, 'weighted_length2'] = net.loc[net[taz_id] == t, 'weighted_length2'] * fac_cell
            net.loc[net[taz_id] == t, 'weight2'] = net.loc[net[taz_id] == t, 'weight2'] *fac_cell
            taz_2.loc[taz_2[taz_id] == t, 'weighted_sub2'] = taz_2.loc[taz_2[taz_id] == t, 'weighted_sub2'] * fac_cell
            taz_2.loc[taz_2[taz_id] == t, 'weight2'] = taz_2.loc[taz_2[taz_id] == t, 'weight2'] * fac_cell

            # list of taz within distance
            t_id = int(taz_2.loc[taz_2[taz_id] == t, taz_mx_id])
            matrix_t = matrix_dis[t_id, :]
            sur_ids = np.where(matrix_t <= distance)[0].tolist()

            # for small taz: add surrounding TAZ within higher distances if no sur found for distance
            if (len(sur_ids)==0) & (float(taz_2.loc[taz_2[taz_id]==t, 'area'])<cell_size):
                sur_ids = np.where(matrix_t == matrix_t.min())[0].tolist()

            # add id of t to sur ids (all relevant cells) and get taz_ids
            sur_ids.append(t_id)
            sur_ids_taz = taz_2.loc[taz_2[taz_mx_id].isin(sur_ids), taz_id].tolist()

            # aggregate weighted network length in sur_ids
            agg_w_len = net.loc[net[taz_id].isin(sur_ids_taz), 'weighted_length2'].sum()
            agg_w_len += taz_2.loc[taz_2[taz_mx_id].isin(sur_ids), 'weighted_sub2'].sum()

            # calculate trips per km for veh_types
            for veh_type in veh_types:
                taz_vkt = float(taz_2.loc[taz_2[taz_id] == t, veh_type])
                taz_veh = taz_vkt / agg_w_len
                net.loc[net[taz_id].isin(sur_ids_taz), '{}_short'.format(veh_type)] = net.loc[net[taz_id].isin(sur_ids_taz), '{}_short'.format(veh_type)] + net.loc[net[taz_id].isin(sur_ids_taz), 'weight2'] * taz_veh
                taz_2.loc[taz_2[taz_id].isin(sur_ids_taz), '{}_sub'.format(veh_type)] += taz_2.loc[taz_2[taz_id].isin(sur_ids_taz), 'weight2'] * taz_veh * taz_2.loc[taz_2[taz_id].isin(sur_ids_taz), sub_len]

        net.drop(columns=['weight', 'weight2', 'weighted_length', 'weighted_length2'], inplace=True)

        cols_keep = ['{}_sub'.format(veh_type) for veh_type in veh_types]
        cols_keep.extend([taz_id])
        taz_2 = taz_2[cols_keep]

        return net, taz_2



# =========================================================
# __init__.py
# @author Nina Thomsen
# @date 21.03.2023
# @copyright Institut fuer Verkehrsforschung,
#            Deutsches Zentrum fuer Luft- und Raumfahrt
# @brief __init__.py file for ULTImodel
# =========================================================

'''import .distribution.DistributeTraffic as DistributeTraffic
import generationData.AttractionFactors as AttractionFactors
import generationData.Matrices as Matrices
import .CreateNetwork as CreateNetwork'''
from .DistributeTraffic import GravityModel
from .DistributeTraffic import IntraZonal
from .DistributeTraffic import TargetValues
from .AttractionFactors import AttractionIndex
from .AttractionFactors import BorderCrossingAtts
from .Matrices import Matrix
from .CreateNetwork import CombineNetworks
from .CreateNetwork import Connectors
from .CreateNetwork import Edges
from .CreateNetwork import Ferries
from .CreateNetwork import Nodes



# =========================================================
# Matrices.py
# @author Nina Thomsen
# @date 11.01.2023
# @copyright Institut fuer Verkehrsforschung,
#            Deutsches Zentrum fuer Luft- und Raumfahrt
# @brief generate travel time and distance matrices between points
# =========================================================

import time
from datetime import datetime

import geopandas as gpd
import numpy as np
import pandas as pd
import requests
from tqdm import tqdm


class Matrix:
    """
    Create cost matrices from OSRM
    """

    def __init__(self, conn, zone_col='zone', id_col='con_id', weight_col='weight', x_='x', y_='y', conn_geo='geometry'):
        """

        :param conn: DF / GDF of connector nodes; if DF, include x and y coordinates
        :param zone_col: name of column with zone ID in conn
        :param id_col: name of connector ID column in conn
        :param weight_col: name of weight column in conn
        :param x_: name of x coordinate column in conn (if conn does not have a geometry column)
        :param y_: name of y coordinate column in conn (if conn does not have a geometry column)
        :param conn_geo: name of geometry column in conn (if conn is a GeoDataFrame)
        :type conn: pd.DataFrame or gpd.GeoDataFrame
        :type zone_col: str
        :type id_col: str
        :type weight_col: str
        :type x_: str
        :type y_: str
        :type conn_geo: str
        """
        self.conn = conn
        if type(self.conn) == gpd.GeoDataFrame:
            self.conn.to_crs(epsg=4326, inplace=True)
            # transform to data frame with x and y coordinates
            self.conn[x_] = self.conn[conn_geo].x
            self.conn[y_] = self.conn[conn_geo].y
            self.conn = self.conn[[id_col, zone_col, weight_col, x_, y_]]

        self.zone_col = zone_col
        self.id_col = id_col
        self.weight_col = weight_col
        self.x_ = x_
        self.y_ = y_

        self.all_mtx = np.zeros((len(conn),len(conn),2))

    def osrm_request_nodes(self, save_np=None):
        """
        OSRM Request: extract travel durations and distances from OSM routing
        
        --> result: self.all_mtx as updated npy array with distances and durations between connector nodes

        :param save_np: default None; filename for saving result matrux, if given, a npy-file of self.all_mtx is saved, updated with the results per iteration
        :type save_np: str
        :return: None
        """
        # create coordinates
        coord = np.c_[self.conn[self.x_], self.conn[self.y_]]
        print("{} connector point coordinates".format(len(coord)))
        str_coords = []
        for i in range(len(coord)):
            coo = np.round(coord[i, :], 6)
            str_ = ','.join(coo.astype(str))
            str_coords.append(str_)

        # OSRM request
        url_str = 'http://router.project-osrm.org/table/v1/driving/{}?sources={}&destinations={}&annotations=duration,distance'

        # base mtx
        self.all_mtx = np.zeros((len(str_coords), len(str_coords), 2))  # from, to, dur/dis

        # 100x100 per request; first 100 rows through rest
        it = int(np.ceil(len(str_coords) / 100))
        st = 0
        time1 = time.time()
        print('start requests:', datetime.now())

        for i in tqdm(range(it)):
            st2 = 0
            en = min(st + 100, len(str_coords))
            starts = ';'.join(np.arange(0, en - st).astype(str))
            s_coords = ';'.join(str_coords[st:en])
            for ii in range(it):
                if i == ii:
                    en2 = en
                    dests = starts
                    coords_url = s_coords
                else:
                    lenstarts = en - st
                    en2 = min(st2 + 100, len(str_coords))
                    dests = ';'.join(np.arange(lenstarts, (en2 - st2) + lenstarts).astype(str))
                    d_coords = ';'.join(str_coords[st2:en2])
                    coords_url = '{};{}'.format(s_coords, d_coords)
                # request
                url_sliced = url_str.format(coords_url, starts, dests)
                r_sl = requests.get(url_sliced)
                # fill matrix
                self.all_mtx[st:en, st2:en2, 0] += np.array(r_sl.json().get('durations')).astype("float64")  # duration
                self.all_mtx[st:en, st2:en2, 1] += np.array(r_sl.json().get('distances')).astype("float64")  # distance
                st2 += 100
            st += 100
            if save_np is not None:
                np.save(save_np, self.all_mtx)
        print('total time requests:', time.time() - time1, 's',
              '({} min)'.format(str(round(float(time.time() - time1) / 60, 2))))
        print(datetime.now())

    def transform_to_taz(self):
        """
        Use connectors weights to generate the mean travel time and distances per taz, based on matrices between connectors
        :return: array with weighted mean travel times and distances and DataFrame with zone IDs and index numbers (index in matrix)
        :rtype: np.array, pd.DataFrame
        """
        # assign conn ID to zones
        self.conn[self.id_col] = np.arange(len(self.conn))
        zones = self.conn[self.zone_col].unique()
        zones_df = pd.DataFrame({"zone": zones}).sort_values("zone")
        zones_df["id"] = np.arange(len(zones_df))

        # weight matrix
        w = np.zeros((len(self.conn), len(self.conn)))

        for a in tqdm(range(len(self.conn))):
            for b in range(len(self.conn)):
                # calc od weight
                w[a, b] = float(self.conn.loc[self.conn[self.id_col] == a, self.weight_col]) * float(self.conn.loc[self.conn[self.id_col] == b, self.weight_col])

        # container for final values
        mx_z = np.zeros((len(zones), len(zones), 2))

        # fill matrix container with weighted mean for zone relations
        time1 = time.time()
        print('start aggregating zone matrix:', datetime.now())
        for o in tqdm(zones):
            id_o = int(zones_df.loc[zones_df['zone'] == o, 'id'])
            id_conn_o = list(self.conn.loc[self.conn[self.zone_col] == o, self.id_col])
            mtx_o = self.all_mtx[id_conn_o, :]
            w_o = w[id_conn_o, :]
            for d in zones:
                id_d = int(zones_df.loc[zones_df['zone'] == d, 'id'])
                id_conn_d = list(self.conn.loc[self.conn[self.zone_col] == d, self.id_col])
                mtx_od = mtx_o[:, id_conn_d]
                w_od = w_o[:, id_conn_d]
                # weighted mean
                mtx_od[:, :, 0] *= w_od
                mtx_od[:, :, 1] *= w_od
                mx_z[id_o, id_d, :] = np.nansum(mtx_od, axis=0).sum(axis=0)  # should be shape (2,)
        print('total time zones:', time.time() - time1, 's',
              '({} min)'.format(str(round(float(time.time() - time1) / 60, 2))))
        print(datetime.now())

        return mx_z, zones_df



# =========================================================
# setup.py
# @author Nina Thomsen
# @date 20.03.2023
# @copyright Institut fuer Verkehrsforschung,
#            Deutsches Zentrum fuer Luft- und Raumfahrt
# @brief setup module for ULTIMO
# =========================================================

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

'''
import required packages from requirements file
with open("requirements.txt") as f:
    INSTALL_REQUIRES = [line.strip() for line in f.readlines()]'''

setuptools.setup(
    name='ultimodel',
    version='1.0.0',
    author='German Aerospace Center - DLR (Nina Thomsen)',
    author_email='nina.thomsen@dlr.de',
    description='Universal transport distribution model',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/DLR-VF/ULTImodel',
    project_urls = {
        "Documentation": 'https://ultimodel.readthedocs.io/',
        "Source": 'https://github.com/DLR-VF/ULTImodel',
        "Bug Tracker": "https://github.com/DLR-VF/ULTImodel/issues "
    },
    license='MIT',
    packages=['ultimodel'],
    install_requires=['pandas', 'numpy', 'geopandas', 'shapely', 'osmnx', 'networkx', 'tqdm', 'requests', 'scikit-learn', 'scipy']
)